import numpy as np
from byzfl.utils.misc import check_vectors_type

class NoAttack(object):

    """
    Description
    -----------
    Class representing an attack which behaves like an honest node.

    Methods
    ---------
    """
    def __init__(self):
        pass 
    
    def __call__(self, honest_vectors):
        """
        Compute the arithmetic mean along axis = 0.
        Returns the average of the array elements.

        Parameters
        ----------
        honest_vectors : 2D ndarray or 2D torch.tensor with floating point or complex dtype
            Matrix containing arrays whose mean is desired.

        Returns
        -------
        mean_vector : ndarray or torch.tensor
            The mean vector of the input. The dtype of the outputs is the same as the input.

        Examples
        --------
        With numpy arrays:
            >>> matrix = np.array([[1., 2., 3.], 
            >>>                    [4., 5., 6.], 
            >>>                    [7., 8., 9.]])
            >>> attack = NoAttack()
            >>> result = attack.get_malicious_vector(matrix)
            >>> print(result)
            ndarray([4. 5. 6.])
        With torch tensors (Warning: We need the tensor to be either a floating point or complex dtype):
            >>> matrix = torch.stack([torch.tensor([1., 2., 3.]),
            >>>                       torch.tensor([4., 5., 6.]), 
            >>>                       torch.tensor([7., 8., 9.])])
            >>> attack = NoAttack()
            >>> result = attack.get_malicious_vector(matrix)
            >>> print(result)
            tensor([4., 5., 6.])
        """

        tools, honest_vectors = check_vectors_type(honest_vectors)
        mean_vector = tools.mean(honest_vectors, axis=0)
        return mean_vector


class SignFlipping:
    
    r"""
    
    Execute the Sign Flipping attack [1]_: send the opposite of the mean vector.

    .. math::

        \mathrm{SignFlipping} \ (x_1, \dots, x_n) = 
        - \frac{1}{n}\sum_{i=1}^{n} x_i    
    
    Initialization parameters
    --------------------------

    None

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors. Conceptually, these vectors correspond to correct gradients submitted by honest workers during a training iteration.

    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output is the same as the input.

    Examples
    --------

        >>> import byzfl
        >>> attack = byzfl.SignFlipping()

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> attack(x)
        array([-4. -5. -6.])
                
        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> attack(x)
        tensor([-4., -5., -6.])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> attack(x)
        array([-4., -5., -6.])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> attack(x)
        tensor([-4., -5., -6.])


    References
    ----------

    .. [1] Zeyuan Allen-Zhu, Faeze Ebrahimianghazani, Jerry Li, and Dan Alistarh. Byzantine-resilient non-convex stochastic gradient descent.
        In International Conference on Learning Representations, 2020

    """

    def __init__(self):
        pass
    
    def __call__(self, honest_vectors):
        tools, honest_vectors = check_vectors_type(honest_vectors)
        mean_vector = tools.mean(honest_vectors, axis=0)
        return tools.multiply(mean_vector, -1)


class FallOfEmpires:
    
    r"""
    Description
    -----------

    Execute the Fall of Empires (FoE) attack [1]_: scale the mean vector by the factor :math:`1 - \tau`.

    .. math::

        \text{FoE}_{\tau}(x_1, \dots, x_n) = 
        (1 - \tau) \cdot \frac{1}{n} \sum_{i=1}^{n} x_i

    where :math:`x_1, \dots, x_n` are the input vectors, and :math:`\tau` is the attack factor.

    Initialization parameters
    --------------------------

    tau : int or float
        The attack factor :math:`\tau` used to adjust the mean vector. Set to 3 by default.

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors. Conceptually, these vectors correspond to correct gradients submitted by honest workers during a training iteration.
    

    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output is the same as the input.
    
    Examples
    --------

    >>> import byzfl
    >>> attack = byzfl.FallofEmpires(3)

    Using numpy arrays
        
    >>> import numpy as np
    >>> x = np.array([[1., 2., 3.],       # np.ndarray
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> attack(x)
    array([ -8. -10. -12.])
            
    Using torch tensors
        
    >>> import torch
    >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
    >>>                   [4., 5., 6.], 
    >>>                   [7., 8., 9.]])
    >>> attack(x)
    tensor([-8., -10., -12.])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> attack(x)
    array([ -8. -10. -12.])

    Using list of torch tensors
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> attack(x)
    tensor([-8., -10., -12.])

    References
    ----------

    .. [1] Cong Xie, Oluwasanmi Koyejo, and Indranil Gupta. Fall of empires: Breaking byzantine-tolerant
            sgd by inner product manipulation. In Ryan P. Adams and Vibhav Gogate (eds.), Proceedings of
            The 35th Uncertainty in Artificial Intelligence Conference, volume 115 of Proceedings of Machine
            Learning Research, pp. 261–270. PMLR, 22–25 Jul 2020. URL https://proceedings.
            mlr.press/v115/xie20a.html.

    """

    def __init__(self, tau=3):
        self.tau = tau

    def __call__(self, honest_vectors):
        tools, honest_vectors = check_vectors_type(honest_vectors)
        mean_vector = tools.mean(honest_vectors, axis=0)
        return tools.multiply(mean_vector, 1 - self.tau)


class ALittleIsEnough():
    
    r"""
    Description
    -----------

    Execute the A Little is Enough (ALIE) attack [1]_: perturb the mean vector using the coordinate-wise standard deviation of the vectors scaled with the attack factor :math:`\tau`.

    .. math::

        \text{ALIE}_{\tau}(x_1, \dots, x_n) = \mu_{x_1, ..., x_n} + \tau \cdot \sigma_{x_1, ..., x_n}
    
    where :math:`\mu_{x_1, \dots, x_n} = \frac{1}{n}\sum_{i=1}^{n}x_i` is the mean vector, :math:`\sigma_{x_1, \dots, x_n} = \sqrt{\frac{1}{n}\sum_{i=1}^{n}(x_i - \mu_{x_1, \dots, x_n})^2}` is the coordinate-wise standard deviation of the input vectors :math:`x_1, \dots, x_n`, and :math:`\tau` is the attack factor.

    Initialization parameters
    --------------------------

    tau : int or float
        The attack factor :math:`\tau` used to adjust the mean vector. Set to 1.5 by default.

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors. Conceptually, these vectors correspond to correct gradients submitted by honest workers during a training iteration.

    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output is the same as the input.
    
    Examples
    --------

    >>> import byzfl
    >>> attack = byzfl.ALittleIsEnough(1.5)

    Using numpy arrays:

    >>> import numppy as np
    >>> x = np.array([[1., 2., 3.], 
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> attack(x)
    array([ 8.5  9.5 10.5])
    
    Using torch tensors (Warning: We need the tensor to be either a floating point or complex dtype):
    
    >>> import torch
    >>> x = torch.stack([torch.tensor([1., 2., 3.]),
    >>>                  torch.tensor([4., 5., 6.]), 
    >>>                  torch.tensor([7., 8., 9.])])
    >>> attack(x)
    tensor([ 8.5000,  9.5000, 10.5000])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> attack(x)
    array([ 8.5  9.5 10.5])

    Using list of torch tensors (Warning: We need the tensor to be either a floating point or complex dtype):
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> attack(x)
    tensor([ 8.5000,  9.5000, 10.5000])
   
    References
    ----------

    .. [1] Baruch, M., Baruch, G., and Goldberg, Y. A little is enough: Circumventing defenses for distributed learning.
           In Advances in Neural Information Processing Systems 32: Annual Conference on Neural Information Processing Systems 2019, 8-14 December 2019, Long Beach, CA, USA, 2019.

    """

    def __init__(self, tau=1.5):
        self.tau = tau
    
    def __call__(self, honest_vectors):
        tools, honest_vectors = check_vectors_type(honest_vectors)
        attack_vector = tools.sqrt(tools.var(honest_vectors, axis=0, ddof=1))
        return tools.add(tools.mean(honest_vectors, axis=0),
                tools.multiply(attack_vector, self.attack_factor))


class Mimic():
    
    r"""
    Description
    -----------

    Execute the Mimic attack [1]_: the attacker mimics the behavior of worker with ID :math:`\epsilon` by sending the same vector as that worker.

    .. math::

        \text{Mimic}_{\epsilon}(x_1, \dots, x_n) = x_{\epsilon}
    
    where :math:`\epsilon \in \{0, 1, \dots, n-1\}` is the ID of the worker to mimic.

    Initialization parameters
    --------------------------

    epsilon : int
        ID of the worker whose behavior is to be mimicked. Set to 0 by default.

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors. Conceptually, these vectors correspond to correct gradients submitted by honest workers during a training iteration.

    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output is the same as the input.
    
    Examples
    --------
    
    >>> import byzfl
    >>> attack = byzfl.Mimic(0)

    Using numpy arrays:

    >>> import numppy as np
    >>> x = np.array([[1., 2., 3.], 
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> attack(x)
    array([1. 2. 3.])

    Using torch tensors:

    >>> import torch
    >>> x = torch.stack([torch.tensor([1., 2., 3.]),
    >>>                  torch.tensor([4., 5., 6.]), 
    >>>                  torch.tensor([7., 8., 9.])])
    >>> attack(x)
    tensor([1., 2., 3.])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> attack(x)
    array([1.  2. 3.])

    Using list of torch tensors:

    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> attack(x)
    tensor([1., 2., 3.])

    References
    ----------

    .. [1] Sai Praneeth Karimireddy, Lie He, and Martin Jaggi. Byzantine-robust learning on heterogeneous datasets via bucketing. In International Conference on Learning Representations, 2022.
    
    """

    def __init__(self, epsilon=0):
        self.epsilon = epsilon
    
    def __call__(self, honest_vectors):
        return honest_vectors[self.epsilon]


class Inf():

    """
    Description
    -----------

    Execute the Infinity attack: generate a vector comprised of positive infinity values.

    Initialization parameters
    --------------------------

    None

    Calling the instance
    --------------------

    Input parameters
    ----------------

    None

    Returns
    -------
    :numpy.ndarray or torch.Tensor

    Examples
    --------

    >>> import byzfl
    >>> attack = byzfl.Inf()

    Using numpy arrays:

    >>> import numppy as np
    >>> x = np.array([[1., 2., 3.], 
    >>>               [4., 5., 6.], 
    >>>               [7., 8., 9.]])
    >>> attack(x)
    array([inf, inf, inf])
    
    Using torch tensors:

    >>> import torch
    >>> x = torch.stack([torch.tensor([1., 2., 3.]),
    >>>                  torch.tensor([4., 5., 6.]), 
    >>>                  torch.tensor([7., 8., 9.])])
    >>> attack(x)
    tensor([inf, inf, inf])

    Using list of numpy arrays

    >>> import numppy as np
    >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
    >>>      np.array([4., 5., 6.]), 
    >>>      np.array([7., 8., 9.])]
    >>> attack(x)
    array([inf, inf, inf])

    Using list of torch tensors (Warning: We need the tensor to be either a floating point or complex dtype):
        
    >>> import torch
    >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
    >>>      torch.tensor([4., 5., 6.]), 
    >>>      torch.tensor([7., 8., 9.])]
    >>> attack(x)
    tensor([inf, inf, inf])

    """ 

    def __init__():
        pass
    
    def __call__(self, honest_vectors):
        tools, honest_vectors = check_vectors_type(honest_vectors)
        return tools.full_like(honest_vectors[0], float('inf'), dtype=np.float64)


from byzfl.pipeline.client import Client
from byzfl.pipeline.byzantine_worker import ByzantineWorker

class ComputeCluster(object):
    """
    Description
    -----------
    This class is the box that contains all the nodes in our
    system (honest and byzantine) and it's responsability is
    manage all this nodes and the information transference.

    Parameters
    -----------
        All this parameters should be passed in a dictionary that contains the following keys.
    nb_workers : int 
        Number of workers
    nb_byz : int
        Number of byzantine nodes
    model-name : str 
        Indicates the model to be used
    device : str 
        Name of the device used
    learning_rate : float 
        Learning rate
    loss_name : str 
        Loss name
    weight_decay : float 
        Regularization used
    milestones : list 
        List of the milestones, where the learning rate decay should be applied
    learning_rate_decay : float 
        Rate decreases over time during training
    momentum : float 
        Momentum
    attack_name : str 
        Name of the attack to be used
    attack_parameters : dict 
        Dictionary with the parameters of the attack where every key is
        the name of the paramater and their value is the value of 
        the parameter
    attack_optimizer_name : (str, optional)
        Name of the optimizer to be used to find the best attack parameters
    attack_optimizer_parameters : (dict, optional)
        Dictionary with the parameters of the optimizer where every key 
        is the name of the paramater and their value is the value of 
        the parameter.
    aggregator_info : dict
        Dictionary with the keys "name" and "parameters" defined.
    pre_agg_list : list 
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.
    dataloaders : list of Dataloader
        List of Dataloader with the train set to every client.
    nb-labels : int
        Number of labels in the dataset

    Methods
    -------
    """
    def __init__(self, params):
        params_client = {
            "model_name": params["model_name"],
            "device": params["device"],
            "learning_rate": params["learning_rate"],
            "loss_name": params["loss_name"],
            "weight_decay": params["weight_decay"],
            "milestones": params["milestones"],
            "learning_rate_decay": params["learning_rate_decay"],
            "attack_name": params["attack_name"],
            "momentum": params["momentum"],
            "nb_labels": params["nb_labels"],
            "nb_workers": params["nb_workers"],
            "nb_steps": params["nb_steps"],
        }

        training_dataloader = params["dataloaders"]

        nb_honest = params["nb_workers"] - params["nb_byz"]

        self.client_list = [
            Client({
                **params_client, 
                "training_dataloader": training_dataloader[idx]
            }) for idx in range(nb_honest)
        ]
        self.labelflipping = False

        if params["attack_name"] == "LabelFlipping":
            self.labelflipping = True
            params["attack_name"] = "NoAttack"
        
        params_byz_worker = {
            "nb_byz": params["nb_byz"],
            "attack_name": params["attack_name"],
            "attack_parameters": params["attack_parameters"],
            "attack_optimizer_name": params["attack_optimizer_name"],
            "attack_optimizer_parameters": params["attack_optimizer_parameters"],
            "aggregator_info": params["aggregator_info"],
            "pre_agg_list": params["pre_agg_list"]
        }

        self.byz_worker = ByzantineWorker(params_byz_worker)
    
    def get_gradients(self):
        """
        Description
        -----------
        Compute and get the gradients of all the clients (Byzantine include)

        Returns
        --------
        Lists of gradients of all clients (honest clients and byzantine)
        """
        [client.compute_gradients() for client in self.client_list]

        honest_gradients = [
            c.get_flat_gradients() 
            for c in self.client_list
        ]

        if self.labelflipping:
            flipped_gradients = [
                c.get_flat_flipped_gradients() 
                for c in self.client_list
            ]
            byzantine_gradients = self.byz_worker.apply_attack(flipped_gradients)
        else:
            byzantine_gradients = self.byz_worker.apply_attack(honest_gradients)

        return honest_gradients + byzantine_gradients
    
    def get_momentum(self):
        """
        Description
        -----------
        Compute and get the momentums of all the clients (Byzantine include)

        Returns
        --------
        Lists of momentums of all clients (honest clients and byzantine)
        """
        [client.compute_gradients() for client in self.client_list]

        honest_momentum = [
            c.get_flat_gradients_with_momentum() 
            for c in self.client_list
        ]
       
        if self.labelflipping:
            flipped_gradients = [
                c.get_flat_flipped_gradients() 
                for c in self.client_list
            ]
            byzantine_gradients = self.byz_worker.apply_attack(flipped_gradients)
        else:
            byzantine_gradients = self.byz_worker.apply_attack(honest_momentum)

        return honest_momentum + byzantine_gradients
    
    def get_batch_norm_stats(self):
        """
        Description
        -----------
        Get the bath-norm stats of all the honest clients

        Returns
        --------
        Lists of bath-norm stats of all honest clients
        """
        batch_norm_stats = [client.get_flat_batch_norm_stats() for client in self.client_list]
        running_mean_stats = [batch_norm_stats_client[0] for batch_norm_stats_client in batch_norm_stats]
        running_var_stats = [batch_norm_stats_client[1] for batch_norm_stats_client in batch_norm_stats]
        byzantine_running_mean = self.byz_worker.apply_batch_norm_attack(running_mean_stats)
        byzantine_running_var = self.byz_worker.apply_batch_norm_attack(running_var_stats)
        return running_mean_stats + byzantine_running_mean, running_var_stats + byzantine_running_var

    
    def set_model_state(self, state_dict):
        """
        Description
        -----------
        Sets the state_dict as the state dict of all the clients.

        Parameters
        ----------
        state_dict(dictionary): state_dict that is desired to be
        set on all the clients.
        """
        [
            client.set_model_state(state_dict) 
            for client in self.client_list
        ]
    
    def transmit_parameters_to_clients(self, parameters):
        """
        Description
        -----------
        Sets the parameters as the parameters of the model of all the clients.

        Parameters
        ----------
        parameters (list): flat vector with the parameters desired to be
        set on all clients.
        """
        [
            client.set_parameters(parameters) 
            for client in self.client_list
        ]

    def get_loss_list_of_clients(self):
        """
        Description
        ------------
        Get the loss list of all clients

        Returns
        -------
        Matrix with the losses that have been computed over the training
        for every client.
        """
        return [client.get_loss_list() for client in self.client_list]
    
    def compute_batch_norm_keys(self):
        """
        Description
        -----------
        Compute batch normalization keys for each client.

        This function iterates over all clients in the Compute Cluster and 
        computes batch normalization keys for each one.

        Returns
        -------
        None
        """
        [client.compute_batch_norm_keys() for client in self.client_list]
    
    def get_train_acc_of_clients(self):
        """
        Description
        -----------
        Retrieve the training accuracy of all clients.

        Returns
        -------
        List
            A list containing the training accuracy for all clients in the Compute Cluster
        """
        return [client.get_train_accuracy() for client in self.client_list]



import torch
import numpy as np

from byzfl.pipeline.model_base_interface import ModelBaseInterface
from byzfl.utils.conversion import flatten_dict

class Client(ModelBaseInterface):
    """
    Description
    -----------
    This class simulates one honest node which is able to train 
    his local model to send the gradients and get the global model every round.

    Parameters
    ----------
        All this parameters should be passed in a dictionary that contains the following keys.
    model_name : str 
        Name of the model to be used
    device : str 
        Name of the device to be used
    learning_rate : float 
        Learning rate
    loss_name : str
        Name of the loss to be used
    weight_decay : float 
        Regularization used
    milestones : list 
        List of the milestones, where the learning rate
        decay should be applied
    learning_rate_decay : float
        Rate decreases over time during training
    attack_name : str 
        Name of the attack to be used
    momentum : float 
        Momentum
    training_dataloader : Dataloader
        Traning dataloader
    nb_labels : int 
        Number of labels in the dataset
    nb_steps : int
        Number of steps in the training

    Methods
    --------
    """
    def __init__(self, params):
        super().__init__({
            "model_name": params["model_name"],
            "device": params["device"],
            "learning_rate": params["learning_rate"],
            "weight_decay": params["weight_decay"],
            "milestones": params["milestones"],
            "learning_rate_decay": params["learning_rate_decay"],
            "nb_workers": params["nb_workers"]
        })

        self.criterion = getattr(torch.nn, params["loss_name"])()

        self.gradient_LF = 0
        self.labelflipping = params["attack_name"] == "LabelFlipping"
        self.nb_labels = params["nb_labels"]
        
        self.momentum = params["momentum"]
        self.momentum_gradient = torch.zeros_like(
            torch.cat(tuple(
                tensor.view(-1) 
                for tensor in self.model.parameters()
            )),
            device=params["device"]
        )
        self.training_dataloader = params["training_dataloader"]
        self.loss_list = np.array([0.0]*params["nb_steps"])
        self.train_acc_list = np.array([0.0]*params["nb_steps"])

        self.SDG_step = 0

    def _sample_train_batch(self):
        """
        Description
        -----------
        Private function to get the next data from the dataloader
        """
        try:
            return next(self.training_dataloader)
        except:
            self.train_iterator = iter(self.training_dataloader)
            return next(self.train_iterator)

    def compute_gradients(self):
        """
        Description
        -----------
        Function where the client compute their gradients 
        of their model loss function.
        """
        inputs, targets = self._sample_train_batch()
        inputs, targets = inputs.to(self.device), targets.to(self.device)

        if self.labelflipping:
            self.model.eval()
            self.model.zero_grad()
            targets_flipped = targets.sub(self.nb_labels - 1).mul(-1)
            outputs = self.model(inputs)
            loss = self.criterion(outputs, targets_flipped)
            loss.backward()
            self.gradient_LF = self.get_dict_gradients()
            self.model.train()
        
        self.model.zero_grad()
        outputs = self.model(inputs)
        loss = self.criterion(outputs, targets)
        self.loss_list[self.SDG_step] = loss.item()
        loss.backward()

        _, predicted = torch.max(outputs.data, 1)
        total = targets.size(0)
        correct = (predicted == targets).sum().item()
        acc = correct/total
        self.train_acc_list[self.SDG_step] = acc

        self.SDG_step += 1
    
    def get_flat_flipped_gradients(self):
        """
        Description
        -----------
        Get the gradients of the model with their targets
        flipped in a flat array.

        Returns
        -------
        List of the gradients.
        """
        return flatten_dict(self.gradient_LF)

    def get_flat_gradients_with_momentum(self):
        """
        Description
        ------------
        Get the gradients of the model applying the momentum 
        in a flat array.

        Returns
        -------
        List with the gradiends with momentum applied.
        """
        self.momentum_gradient.mul_(self.momentum)
        self.momentum_gradient.add_(
            self.get_flat_gradients(),
            alpha=1-self.momentum
        )

        return self.momentum_gradient
    
    def get_loss_list(self):
        """
        Description
        ------------
        Get the loss list of the client

        Returns
        -------
        List with the losses that have been computed over the training.
        """
        return self.loss_list
    
    def get_train_accuracy(self):
        """
        Description
        ------------
        Get the batch train accuracy list of the client

        Returns
        -------
        List with the accuracies that have been computed over the training.
        """
        return self.train_acc_list
    
    def set_model_state(self, state_dict):
        """
        Description
        -----------
        Sets the state_dict of the model for the state_dict given by parameter.

        Parameters
        ----------
        state_dict : dict 
            State_dict from a model
        """
        self.model.load_state_dict(state_dict)



import torch, torchvision, random
import torchvision.transforms as T
import numpy as np
from torch.utils.data import DataLoader, random_split

transforms_hflip = T.Compose([T.RandomHorizontalFlip(), T.ToTensor()])
transforms_mnist = T.Compose([T.ToTensor(), T.Normalize((0.1307,), (0.3081,))])
transforms_cifar_train = T.Compose([
        T.RandomCrop(32, padding=4),
        T.RandomHorizontalFlip(),
        T.ToTensor(),
        T.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])
transforms_cifar_test = T.Compose([
    T.ToTensor(),
    T.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
])

dict_ = {
    "mnist":        ["MNIST", transforms_mnist, transforms_mnist],
    "fashionmnist": ["FashionMNIST", transforms_hflip, transforms_hflip],
    "emnist":       ["EMNIST", transforms_mnist, transforms_mnist],
    "cifar10":      ["CIFAR10", transforms_cifar_train, transforms_cifar_test],
    "cifar100":     ["CIFAR100", transforms_cifar_train, transforms_cifar_test],
    "imagenet":     ["ImageNet", transforms_hflip, transforms_hflip]
}

def get_dataloaders(params):

    dataset_name = params["dataset_name"]
    dataset = getattr(torchvision.datasets, dict_[dataset_name][0])(
            root = params["data_folder"], 
            train = True, 
            download = True,
            transform = dict_[dataset_name][1]
    )
    
    dataset.targets = torch.Tensor(dataset.targets).long()
    train_size = int(params["size_train_set"] * len(dataset))
    val_size = len(dataset) - train_size

    train_dataset, val_dataset = random_split(dataset, [train_size, val_size])

    validation_dataloader = None

    if len(val_dataset) > 0:

        validation_dataloader = DataLoader(val_dataset, 
                                        batch_size=params["batch_size_validation"], 
                                        shuffle=True)
        
    else:
        print("WARNING: NO VALIDATION DATASET")

    nb_honest = params["nb_workers"] - params["nb_byz"]
    
    params_split_datasets = {
        "nb_honest": nb_honest,
        "data_distribution_name": params["data_distribution_name"],
        "distribution_parameter": params["distribution_parameter"],
        "batch_size": params["batch_size"],
        "dataset": train_dataset
    }

    training_dataloaders = split_datasets(params_split_datasets)

    test_dataset = getattr(torchvision.datasets, dict_[dataset_name][0])(
                root = params["data_folder"],
                train=False, 
                download=True,
                transform=dict_[dataset_name][1]
    )

    test_dataloader = DataLoader(test_dataset, batch_size=params["batch_size_validation"], shuffle=True)

    return training_dataloaders, validation_dataloader, test_dataloader

def split_datasets(params):
    data_dist = params["data_distribution_name"]
    distribution_parameter = params["distribution_parameter"]
    nb_honest = params["nb_honest"] 
    dataset = params["dataset"]
    #targets = dataset.dataset.targets[dataset.indices]
    targets = dataset.dataset.targets
    idx = dataset.indices

    match data_dist:
        case 'iid':
            split_idx = iid_idx(idx, nb_honest)
        case 'gamma_similarity_niid':
            split_idx = gamma_niid_idx(targets, idx, nb_honest, distribution_parameter)
        case 'dirichlet_niid':
            split_idx = dirichlet_niid_idx(targets, idx, nb_honest, distribution_parameter)
        case 'extreme_niid':
            split_idx = extreme_niid_idx(targets, idx, nb_honest)
        case _:
            raise ValueError(f"Invalid value for data_dist: {data_dist}")
    
    return idx_to_dataloaders(dataset, split_idx, params["batch_size"])

def iid_idx(idx, nb_honest):
    random.shuffle(idx)
    split_idx = np.array_split(idx, nb_honest)
    return split_idx

def extreme_niid_idx(targets, idx, nb_honest):
    if len(idx) == 0:
        return list([[]]*nb_honest)
    sorted_idx = np.array(sorted(zip(targets[idx],idx)))[:,1]
    split_idx = np.array_split(sorted_idx, nb_honest)
    return split_idx

def gamma_niid_idx(targets, idx, nb_honest, gamma):
    nb_similarity = int(len(idx)*gamma)
    iid = iid_idx(idx[:nb_similarity], nb_honest)
    niid = extreme_niid_idx(targets, idx[nb_similarity:], nb_honest)
    split_idx = [np.concatenate((iid[i],niid[i])) for i in range(nb_honest)]
    split_idx = [node_idx.astype(int) for node_idx in split_idx]
    return split_idx

def dirichlet_niid_idx(targets, idx, nb_honest, alpha):
    c = len(torch.unique(targets))
    sample = np.random.dirichlet(np.repeat(alpha, nb_honest), size=c)
    p = np.cumsum(sample, axis=1)[:,:-1]
    aux_idx = [np.where(targets[idx] == k)[0] for k in range(c)]
    aux_idx = [np.split(aux_idx[k], (p[k]*len(aux_idx[k])).astype(int)) for k in range(c)]
    aux_idx = [np.concatenate([aux_idx[i][j] for i in range(c)]) for j in range(nb_honest)]
    idx = np.array(idx)
    aux_idx = [list(idx[aux_idx[i]]) for i in range(len(aux_idx))]
    return aux_idx

def idx_to_dataloaders(dataset, split_idx, batch_size):
    data_loaders = []
    for i in range(len(split_idx)):
        subset = torch.utils.data.Subset(dataset.dataset, split_idx[i])
        data_loader = DataLoader(subset, batch_size=batch_size, shuffle=True)
        data_loaders.append(data_loader)
    return data_loaders



import random
import time

import numpy as np
import torch

from byzfl.pipeline.server import Server
from byzfl.pipeline.compute_cluster import ComputeCluster
from byzfl.pipeline.dataset import get_dataloaders
from byzfl.pipeline.managers import FileManager, ParamsManager

class Train(object):
    """
    Description
    -----------
    Class with implements the train algorithms and stores some stadistics
    of the training

    Parameters
    ----------
        All this parameters should be passed in a dictionary that contains the following keys.
    nb_steps : int 
        Number of steps to do in the Training
    evaluation_delta : int
        How many steps it have to wait to compute accuracy
    evaulate_on_test : bool
        Indicates if will be evaluate on test set
    model_name : str
        Indicates the model to be used
    device : str 
        Name of the device used
    bit_precision : int
        How many bits will be displayed in the accuracy
    learning_rate : float
        Learning rate
    weight_decay : float 
        Weight decay used
    learning_rate_decay : float
        Lerning rate decay used
    batch_size : int
        Batch size used in the train dataloaders
    batch_size_validation: int
        Batch size used in the validation and test dataloader
    size_train_set: float
        Proportion of data from the train that will be used to train
    dataset_name : str 
        Name of the dataset used
    nb_labels : int
        Number of labels in the dataset
    data_folder : str
        Path of the folder that contains the data
    data_distribution_name : str
        Name of the data distribution
    distribution_parameter : float
        Parameter for the data distribution
    aggregator_info : dict
        Dictionary with the keys "name" and "parameters" defined.
    pre_agg_list : list
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.
    nb_workers : int
        Number of workers
    nb_byz : int
        Number of byzantine nodes
    declared_nb_byz : int
        Number of byzantine nodes that the server will try to defend against
    loss_name : str 
        Loss name to be used
    milestones : list 
        List of the milestones, where the learning rate decay should be applied
    learning_rate_decay : float
        Rate decreases over time during training
    momentum : float
        Momentum
    batch_norm_momentum : float
        Momentum used in the federated batch norm
    attack_name : str 
        Attack name
    attack_parameters : dict 
        Dictionary with the parameters of the attack where every key is the name 
        of the paramater and their value is the value of the parameter
    attack_optimizer_name : (str, optional)
        Name of the optimizer to be used
    attack_optimizer_parameters : (dict, optional)
        Dictionary with the parameters of the optimizer where every 
        key is the name of the paramater and their value is the value 
        of the parameter
    store_models : (bool)
        If true every delta steps the model will be saved

    Methods
    -------
    """
    def __init__(self, raw_params):

        params_manager = ParamsManager(params=raw_params)
        params = params_manager.get_flatten_info()

        if params["nb_honest"] is not None:
            params["nb_workers"] = params["nb_honest"] + params["declared_nb_byz"]

        file_manager_params = {
            "result_path": params["results_directory"],
            "dataset_name": params["dataset_name"],
            "model_name": params["model_name"],
            "nb_workers": params["nb_workers"],
            "nb_byz": params["nb_byz"],
            "declared_nb_byz": params["declared_nb_byz"],
            "data_distribution_name": params["data_distribution_name"],
            "distribution_parameter": params["distribution_parameter"],
            "aggregation_name": params["aggregator_info"]["name"],
            "pre_aggregation_names":  [
                dict['name']
                for dict in params["pre_agg_list"]
            ],
            "attack_name": params["attack_name"],
            "learning_rate": params["learning_rate"],
            "momentum": params["momentum"],
            "weight_decay": params["weight_decay"],
            "learning_rate_decay": params["learning_rate_decay"]
        }

        self.file_manager = FileManager(file_manager_params)
        self.file_manager.save_config_dict(params_manager.get_data())

        params["aggregator_info"]["parameters"]["f"] = params["declared_nb_byz"]
        if len(params["pre_agg_list"]) > 0:
            for pre_agg in params["pre_agg_list"]:
                pre_agg["parameters"]["f"] = params["declared_nb_byz"]

        params_dataloaders = {
            "dataset_name": params["dataset_name"],
            "batch_size": params["batch_size"],
            "batch_size_validation": params["batch_size_validation"],
            "size_train_set": params["size_train_set"],
            "dataset_name": params["dataset_name"],
            "nb_labels": params["nb_labels"],
            "data_folder": params["data_folder"],
            "data_distribution_name": params["data_distribution_name"],
            "distribution_parameter": params["distribution_parameter"],
            "nb_workers": params["nb_workers"],
            "nb_byz": params["nb_byz"],
        }

        self.data_distribution_seed = params["data_distribution_seed"]
        
        # Deterministic
        # https://pytorch.org/docs/stable/notes/randomness.html
        np.random.seed(self.data_distribution_seed)
        torch.manual_seed(self.data_distribution_seed)
        torch.use_deterministic_algorithms(True)
        random.seed(self.data_distribution_seed)

        train_dataloaders, self.validation_dataloader, \
        self.test_dataloader = get_dataloaders(params_dataloaders)

        self.use_validation = self.validation_dataloader != None

        self.training_seed = params["training_seed"]
        np.random.seed(self.training_seed)
        torch.manual_seed(self.training_seed)
        random.seed(self.training_seed)

        server_params = {
            "model_name": params["model_name"],
            "nb_workers": params["nb_workers"],
            "evaluate_on_test": params["evaluate_on_test"],
            "device": params["device"],
            "learning_rate": params["learning_rate"],
            "weight_decay": params["weight_decay"],
            "milestones": params["milestones"],
            "learning_rate_decay": params["learning_rate_decay"],
            "aggregator_info": params["aggregator_info"],
            "pre_agg_list": params["pre_agg_list"],
            "batch_size": params["batch_size"],
            "batch_norm_momentum": params["batch_norm_momentum"]
        }

        self.server = Server(server_params)

        if params["nb_byz"] == 0:
            params["attack_name"] = "NoAttack"

        compute_cluster_params = {
            "device": params["device"],
            "nb_workers": params["nb_workers"],
            "nb_byz": params["nb_byz"],
            "model_name": params["model_name"],
            "learning_rate": params["learning_rate"],
            "loss_name": params["loss_name"],
            "weight_decay": params["weight_decay"],
            "milestones": params["milestones"],
            "learning_rate_decay": params["learning_rate_decay"],
            "momentum": params["momentum"],
            "attack_name": params["attack_name"],
            "attack_parameters": params["attack_parameters"],
            "attack_optimizer_name": params["attack_optimizer_name"],
            "attack_optimizer_parameters": params["attack_optimizer_parameters"],
            "aggregator_info": params["aggregator_info"],
            "pre_agg_list": params["pre_agg_list"],
            "dataloaders": train_dataloaders,
            "nb_labels": params["nb_labels"],
            "nb_steps": params["nb_steps"]
        }

        self.compute_cluster = ComputeCluster(compute_cluster_params)

        self.compute_cluster.set_model_state(self.server.get_dict_parameters())

        self.device = params["device"]
        self.steps = params["nb_steps"]
        self.evaluation_delta = params["evaluation_delta"]
        self.evaluate_on_test = params["evaluate_on_test"]
        self.store_training_accuracy = params["store_training_accuracy"]
        self.store_training_loss = params["store_training_loss"]
        self.store_models = params["store_models"]

        #Stored for display results only
        self.agg_name = params["aggregator_info"]["name"]
        self.data_dist_name = params["data_distribution_name"]
        self.attack_name = params["attack_name"]
        self.nb_byz = params["nb_byz"]

        self.accuracy_list = np.array([])
        self.test_accuracy_list = np.array([])
        self.loss_list = np.array([])

        self.use_batch_norm_stats = False
    
    @torch.no_grad()
    def _compute_accuracy(self, model, dataloader):
        """
        Description
        -----------
        Compute the accuracy using the test set of the model

        Returns
        -------
        A float with the accuracy value
        """
        total = 0
        correct = 0
        for inputs, targets in dataloader:
            inputs, targets = inputs.to(self.device), targets.to(self.device)
            outputs = model(inputs)
            _, predicted = torch.max(outputs.data, 1)
            total += targets.size(0)
            correct += (predicted == targets).sum().item()
        return correct/total

    def run_SGD(self):
        """
        Description
        -----------
        Trains the model running SGD for n steps (setting.json)
        """
        start_time = time.time()

        self.server.compute_batch_norm_keys()

        if self.server.use_batch_norm():
            self.use_batch_norm_stats = True
            self.compute_cluster.compute_batch_norm_keys()
        
        for step in range(0, self.steps):

            if step % self.evaluation_delta == 0:
                model = self.server.get_model()
                if self.use_validation:
                    accuracy = self._compute_accuracy(model, self.validation_dataloader)
                    self.accuracy_list = np.append(self.accuracy_list, accuracy)

                if self.evaluate_on_test:
                    test_accuracy = self._compute_accuracy(model, self.test_dataloader)
                    self.test_accuracy_list = np.append(
                        self.test_accuracy_list, 
                        test_accuracy
                    )
                    self.file_manager.write_array_in_file(
                        self.test_accuracy_list, 
                        "test_accuracy_tr_seed_" + str(self.training_seed) 
                        + "_dd_seed_" + str(self.data_distribution_seed) +".txt"
                    )
                
                if self.store_models:
                    self.file_manager.save_state_dict(
                        self.server.get_dict_parameters(),
                        self.training_seed,
                        step
                    )
            
            #Training
            new_gradients = self.compute_cluster.get_momentum()

            if self.use_batch_norm_stats:
                new_running_mean, new_running_var = self.compute_cluster.get_batch_norm_stats()

            #Aggregation and update of the global model
            self.server.update_model(new_gradients)

            if self.use_batch_norm_stats:
                self.server.update_batch_norm_stats(new_running_mean, new_running_var)

            #Broadcasting
            new_parameters = self.server.get_dict_parameters()
            self.compute_cluster.set_model_state(new_parameters)
        

        self.loss_list = self.compute_cluster.get_loss_list_of_clients()
        self.train_accuracy_list = self.compute_cluster.get_train_acc_of_clients()

        if self.use_validation:
            accuracy = self._compute_accuracy(model, self.validation_dataloader)
            self.accuracy_list = np.append(self.accuracy_list, accuracy)
            self.file_manager.write_array_in_file(
                self.accuracy_list, 
                "validation_accuracy_tr_seed_" + str(self.training_seed) 
                + "_dd_seed_" + str(self.data_distribution_seed) +".txt"
            )
        
        if self.evaluate_on_test:
            test_accuracy = self._compute_accuracy(model, self.test_dataloader)
            self.test_accuracy_list = np.append(self.test_accuracy_list, test_accuracy)
            self.file_manager.write_array_in_file(
                self.test_accuracy_list, 
                "test_accuracy_tr_seed_" + str(self.training_seed) 
                + "_dd_seed_" + str(self.data_distribution_seed) +".txt"
            )

        if self.store_training_loss:
            for i, loss in enumerate(self.loss_list):
                self.file_manager.save_loss(
                    loss, 
                    self.training_seed, 
                    self.data_distribution_seed, 
                    i
                )
        
        if self.store_training_accuracy:
            for i, acc in enumerate(self.train_accuracy_list):
                self.file_manager.save_accuracy(
                    acc, 
                    self.training_seed, 
                    self.data_distribution_seed,
                    i
                )
        
        if self.store_models:
            self.file_manager.save_state_dict(
                self.server.get_dict_parameters(),
                self.training_seed,
                self.steps
            )
        
        end_time = time.time()
        execution_time = end_time - start_time

        self.file_manager.write_array_in_file(
            np.array(execution_time),
            "train_time_tr_seed_" + str(self.training_seed) 
            + "_dd_seed_" + str(self.data_distribution_seed) +".txt"
        )



import numpy as np
import torch
from scipy.spatial import distance

from byzfl.utils import torch_tools


def check_vectors_type(vectors):
    if isinstance(vectors, list) and isinstance(vectors[0], np.ndarray):
        tools = np
        vectors = np.array(vectors)
    elif isinstance(vectors, np.ndarray):
        tools = np
    elif isinstance(vectors, list) and isinstance(vectors[0], torch.Tensor):
        tools = torch_tools
        vectors = torch.stack(vectors).float()
    elif isinstance(vectors, torch.Tensor):
        tools = torch_tools
        vectors = vectors.float()
    else :
        raise TypeError("'vectors' should be a 'list' of"
                        " ('np.ndarray' or 'torch.Tensor') or 'np.ndarray'"
                        " or 'torch.Tensor'")
    return tools, vectors

def random_tool(vectors):
    if (isinstance(vectors, list) and isinstance(vectors[0], np.ndarray) or
        isinstance(vectors, np.ndarray)):
        tools = np.random
    elif (isinstance(vectors, list) and isinstance(vectors[0], torch.Tensor) or
          isinstance(vectors, torch.Tensor)):
        tools = torch_tools

    return tools

def distance_tool(vectors):
    if (isinstance(vectors, list) and isinstance(vectors[0], np.ndarray) or
        isinstance(vectors, np.ndarray)):
        tools = distance
    elif (isinstance(vectors, list) and isinstance(vectors[0], torch.Tensor) or
          isinstance(vectors, torch.Tensor)):
        tools = torch_tools

    return tools


def check_type(element, t):
    if isinstance(t, tuple):
        s = "" 
        for i in range(len(t)):
            s = s + "'"+ t[i].__name__+"'"
            if i < len(t)-1:
                s = s + " or "
    else:
        s = "'"+t.__name__+"'"
    if not isinstance(element, t):
        raise TypeError("Expected "+s+" but got '"+type(element).__name__+"'")



import math
import json
import os

import matplotlib.pyplot as plt
import numpy as np
from numpy import genfromtxt
import seaborn as sns

class Plot(object):
    
    def custom_dict_to_str(self, d):
        if not d:
            return ''
        else:
            return str(d)
    
    def plot_accuracy_fix_agg(self, path_to_results, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        tab_sign = ['-', '--', '-.', ':', 'solid']
        markers = ['^','s', '<', 'o', '*']

        colors = [(0, 0.4470, 0.7410), (0.8500, 0.3250, 0.0980), (0.4660, 0.6740, 0.1880), (120/255,120/255, 120/255), (0.7, 0.2, 0.5)]
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_list = data["honest_nodes"]["learning_rate"]
        momentum_list = data["honest_nodes"]["momentum"]
        wd_list = data["honest_nodes"]["weight_decay"]
        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_list, float):
            lr_list = [lr_list]
        if isinstance(momentum_list, float):
            momentum_list = [momentum_list]
        if isinstance(wd_list, float):
            wd_list = [wd_list]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                for lr in lr_list:
                    for momentum in momentum_list:
                        for wd in wd_list:
                            for lr_d in lr_decay:
                                for data_dist in data_distributions:
                                    alpha = ""
                                    if data_dist["name"] == "dirichlet_niid":
                                        alpha = data_dist["parameters"]["alpha"]
                                    for agg in aggregators:
                                        for pre_agg in pre_aggregators:
                                            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                                            pre_agg_names = "_".join(pre_agg_list_names)
                                            tab_acc = np.zeros((
                                                len(attacks), 
                                                nb_seeds, 
                                                nb_accuracies
                                            ))
                                            for i, attack in enumerate(attacks):
                                                for run in range(nb_seeds):
                                                    common_name = str(dataset_name + "_" 
                                                                    + model_name + "_n_" 
                                                                    + str(nb_nodes) + "_f_" 
                                                                    + str(nb_byzantine) + "_" 
                                                                    + self.custom_dict_to_str(data_dist["name"]) 
                                                                    + str(alpha) + "_" 
                                                                    + self.custom_dict_to_str(agg["name"]) + "_" 
                                                                    + pre_agg_names + "_" 
                                                                    + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                                    + str(lr) + "_mom_"
                                                                    + str(momentum) + "_wd_"
                                                                    + str(wd)+ "_lr_decay_"
                                                                    + str(lr_d))
                                                    tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                            
                                            err = np.zeros((len(attacks), nb_accuracies))
                                            for i in range(len(err)):
                                                err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                            
                                            plt.rcParams.update({'font.size': 12})

                                            
                                            for i, attack in enumerate(attacks):
                                                attack = attack["name"]
                                                plt.plot(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0), label = attack, color = colors[i], linestyle = tab_sign[i], marker = markers[i], markevery = 1)
                                                plt.fill_between(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0) - err[i], np.mean(tab_acc[i], axis = 0) + err[i], alpha = 0.25)

                                            plt.xlabel('Round')
                                            plt.ylabel('Accuracy')
                                            plt.xlim(0,(nb_accuracies-1)*evaluation_delta)
                                            plt.ylim(0,1)
                                            plt.grid()
                                            plt.legend()

                                            plot_name = str(dataset_name + "_"
                                                            + model_name + "_n_"
                                                            + str(nb_nodes) + "_f_"
                                                            + str(nb_byzantine) + "_"
                                                            + self.custom_dict_to_str(data_dist["name"])
                                                            + str(alpha) + "_"
                                                            + self.custom_dict_to_str(agg["name"]) + "_"
                                                            + pre_agg_names + "_lr_"
                                                            + str(lr) + "_mom_"
                                                            + str(momentum) + "_wd_"
                                                            + str(wd)+ "_lr_decay_"
                                                            + str(lr_d))
                                            plt.savefig(path_to_plot+"/"+plot_name+'_plot.pdf')
                                            plt.close()
    
    def plot_accuracy_fix_attack(self, path_to_results, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        tab_sign = ['-', '--', '-.', ':', 'solid']
        markers = ['^','s', '<', 'o', '*']

        colors = [(0, 0.4470, 0.7410), (0.8500, 0.3250, 0.0980), (0.4660, 0.6740, 0.1880), (120/255,120/255, 120/255), (0.7, 0.2, 0.5)]
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_list = data["honest_nodes"]["learning_rate"]
        momentum_list = data["honest_nodes"]["momentum"]
        wd_list = data["honest_nodes"]["weight_decay"]
        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_list, float):
            lr_list = [lr_list]
        if isinstance(momentum_list, float):
            momentum_list = [momentum_list]
        if isinstance(wd_list, float):
            wd_list = [wd_list]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                for lr in lr_list:
                    for momentum in momentum_list:
                        for wd in wd_list:
                            for lr_d in lr_decay:
                                for data_dist in data_distributions:
                                    alpha = ""
                                    if data_dist["name"] == "dirichlet_niid":
                                        alpha = data_dist["parameters"]["alpha"]
                                    for attack in attacks:
                                        for pre_agg in pre_aggregators:
                                            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                                            pre_agg_names = "_".join(pre_agg_list_names)
                                            tab_acc = np.zeros((
                                                len(aggregators), 
                                                nb_seeds, 
                                                nb_accuracies
                                            ))
                                            for i, agg in enumerate(aggregators):
                                                for run in range(nb_seeds):
                                                    common_name = str(dataset_name + "_" 
                                                                    + model_name + "_n_" 
                                                                    + str(nb_nodes) + "_f_" 
                                                                    + str(nb_byzantine) + "_" 
                                                                    + self.custom_dict_to_str(data_dist["name"]) 
                                                                    + str(alpha) + "_" 
                                                                    + self.custom_dict_to_str(agg["name"]) + "_" 
                                                                    + pre_agg_names + "_" 
                                                                    + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                                    + str(lr) + "_mom_"
                                                                    + str(momentum) + "_wd_"
                                                                    + str(wd)+ "_lr_decay_"
                                                                    + str(lr_d))
                                                    tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                            
                                            err = np.zeros((len(aggregators), nb_accuracies))
                                            for i in range(len(err)):
                                                err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                            
                                            plt.rcParams.update({'font.size': 12})

                                            
                                            for i, agg in enumerate(aggregators):
                                                agg = agg["name"]
                                                plt.plot(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0), label = agg, color = colors[i], linestyle = tab_sign[i], marker = markers[i], markevery = 1)
                                                plt.fill_between(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0) - err[i], np.mean(tab_acc[i], axis = 0) + err[i], alpha = 0.25)

                                            plt.xlabel('Round')
                                            plt.ylabel('Accuracy')
                                            plt.xlim(0,(nb_accuracies-1)*evaluation_delta)
                                            plt.ylim(0,1)
                                            plt.grid()
                                            plt.legend()

                                            plot_name = str(dataset_name + "_"
                                                            + model_name + "_n_"
                                                            + str(nb_nodes) + "_f_"
                                                            + str(nb_byzantine) + "_"
                                                            + self.custom_dict_to_str(data_dist["name"])
                                                            + str(alpha) + "_"
                                                            + str(attack["name"]) + "_"
                                                            + pre_agg_names + "_lr_"
                                                            + str(lr) + "_mom_"
                                                            + str(momentum) + "_wd_"
                                                            + str(wd)+ "_lr_decay_"
                                                            + str(lr_d))
                                            plt.savefig(path_to_plot+"/"+plot_name+'_plot.pdf')
                                            plt.close()
    
    def plot_minimum_of_max_accuracies(self, path_to_results, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        tab_sign = ['-', '--', '-.', ':', 'solid']
        markers = ['^','s', '<', 'o', '*']

        colors = [(0, 0.4470, 0.7410), (0.8500, 0.3250, 0.0980), (0.4660, 0.6740, 0.1880), (0.5, 0.5, 0.5), (0.7, 0.2, 0.5)]
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_list = data["honest_nodes"]["learning_rate"]
        momentum_list = data["honest_nodes"]["momentum"]
        wd_list = data["honest_nodes"]["weight_decay"]
        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_list, float):
            lr_list = [lr_list]
        if isinstance(momentum_list, float):
            momentum_list = [momentum_list]
        if isinstance(wd_list, float):
            wd_list = [wd_list]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                for lr in lr_list:
                    for momentum in momentum_list:
                        for wd in wd_list:
                            for lr_d in lr_decay:
                                for data_dist in data_distributions:
                                    alpha = ""
                                    if data_dist["name"] == "dirichlet_niid":
                                        alpha = data_dist["parameters"]["alpha"]
                                    for pre_agg in pre_aggregators:
                                        pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                                        pre_agg_names = "_".join(pre_agg_list_names)
                                        minimum_max = np.array([np.inf]*len(aggregators))
                                        tab_acc_min_max = np.zeros((
                                                len(aggregators),
                                                nb_accuracies
                                        ))
                                        tab_err_min_max = np.zeros((
                                                len(aggregators),
                                                nb_accuracies
                                        ))
                                        for attack in attacks:
                                            tab_acc = np.zeros((
                                                len(aggregators), 
                                                nb_seeds, 
                                                nb_accuracies
                                            ))
                                            for i, agg in enumerate(aggregators):
                                                for run in range(nb_seeds):
                                                    common_name = str(dataset_name + "_" 
                                                                    + model_name + "_n_" 
                                                                    + str(nb_nodes) + "_f_" 
                                                                    + str(nb_byzantine) + "_" 
                                                                    + self.custom_dict_to_str(data_dist["name"]) 
                                                                    + str(alpha) + "_" 
                                                                    + self.custom_dict_to_str(agg["name"]) + "_" 
                                                                    + pre_agg_names + "_" 
                                                                    + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                                    + str(lr) + "_mom_"
                                                                    + str(momentum) + "_wd_"
                                                                    + str(wd)+ "_lr_decay_"
                                                                    + str(lr_d))
                                                    tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                            
                                            err = np.zeros((len(aggregators), nb_accuracies))
                                            for i in range(len(err)):
                                                err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                            
                                            plt.rcParams.update({'font.size': 12})

                                            for i, agg in enumerate(aggregators):
                                                mean_accuraccy_agg = np.mean(tab_acc[i], axis = 0)
                                                max_value = np.max(mean_accuraccy_agg)
                                                if max_value < minimum_max[i]:
                                                    minimum_max[i] = max_value
                                                    tab_acc_min_max[i] = mean_accuraccy_agg
                                                    tab_err_min_max[i] = err[i]

                                        plt.rcParams.update({'font.size': 12})

                                            
                                        for i, agg in enumerate(aggregators):
                                            agg = agg["name"]
                                            plt.plot(np.arange(nb_accuracies)*evaluation_delta, tab_acc_min_max[i], label = agg, color = colors[i], linestyle = tab_sign[i], marker = markers[i], markevery = 1, alpha=0.8)
                                            plt.fill_between(np.arange(nb_accuracies)*evaluation_delta, tab_acc_min_max[i] - tab_err_min_max[i], tab_acc_min_max[i] + tab_err_min_max[i], alpha = 0.25)

                                        plt.xlabel('Round')
                                        plt.ylabel('Accuracy')
                                        plt.xlim(0,(nb_accuracies-1)*evaluation_delta)
                                        plt.ylim(0,1)
                                        plt.grid()
                                        plt.legend()

                                        plot_name = str(dataset_name + "_"
                                                        + model_name + "_n_"
                                                        + str(nb_nodes) + "_f_"
                                                        + str(nb_byzantine) + "_"
                                                        + self.custom_dict_to_str(data_dist["name"])
                                                        + str(alpha) + "_"
                                                        + pre_agg_names + "_lr_"
                                                        + str(lr) + "_mom_"
                                                        + str(momentum) + "_wd_"
                                                        + str(wd)+ "_lr_decay_"
                                                        + str(lr_d))
                                        plt.savefig(path_to_plot+"/"+plot_name+'_plot.pdf')
                                        plt.close()
    
    def plot_accuracy_fix_agg_best_setting(self, path_to_results, path_to_plot, path_to_hyperparameters):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        tab_sign = ['-', '--', '-.', ':', 'solid']
        markers = ['^','s', '<', 'o', '*']

        colors = [(0, 0.4470, 0.7410), (0.8500, 0.3250, 0.0980), (0.4660, 0.6740, 0.1880), (120/255,120/255, 120/255), (0.7, 0.2, 0.5)]
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for nb_nodes in nb_workers:
            for nb_byzantine in nb_byz:
                if change:
                    nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                for lr_d in lr_decay:
                    for data_dist in data_distributions:
                        alpha_list = [""]
                        if data_dist["name"] == "dirichlet_niid":
                            alpha_list = data_dist["parameters"]["alpha"]
                        if data_dist["name"] == "gamma_similarity_niid":
                            alpha_list = data_dist["parameters"]["gamma"]
                        for alpha in alpha_list:
                            for pre_agg in pre_aggregators:
                                pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
                                pre_agg_names = "_".join(pre_agg_list_names)
                                for agg in aggregators:
                                    hyperparameters_file_name = str(dataset_name + "_"
                                                                    + model_name + "_n_"
                                                                    + str(nb_nodes) + "_f_"
                                                                    + str(nb_byzantine) + "_"
                                                                    + self.custom_dict_to_str(data_dist["name"])
                                                                    + str(alpha) + "_"
                                                                    + pre_agg_names + "_"
                                                                    + agg["name"]
                                                                    + ".txt")
                                    
                                    hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                    lr = hyperparameters[0]
                                    momentum = hyperparameters[1]
                                    if momentum == 0.0:
                                        momentum = 0
                                    wd = hyperparameters[2]

                                    tab_acc = np.zeros((
                                        len(attacks), 
                                        nb_seeds, 
                                        nb_accuracies
                                    ))
                                    for i, attack in enumerate(attacks):
                                        for run in range(nb_seeds):
                                            common_name = str(dataset_name + "_" 
                                                            + model_name + "_n_" 
                                                            + str(nb_nodes) + "_f_" 
                                                            + str(nb_byzantine) + "_" 
                                                            + self.custom_dict_to_str(data_dist["name"]) 
                                                            + str(alpha) + "_" 
                                                            + self.custom_dict_to_str(agg["name"]) + "_" 
                                                            + pre_agg_names + "_" 
                                                            + self.custom_dict_to_str(attack["name"]) + "_lr_"
                                                            + str(lr) + "_mom_"
                                                            + str(momentum) + "_wd_"
                                                            + str(wd)+ "_lr_decay_"
                                                            + str(lr_d))
                                            tab_acc[i][run] = genfromtxt(path_to_results+"/"+common_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                    
                                    err = np.zeros((len(attacks), nb_accuracies))
                                    for i in range(len(err)):
                                        err[i] = (1.96*np.std(tab_acc[i], axis = 0))/math.sqrt(nb_seeds)
                                    
                                    plt.rcParams.update({'font.size': 12})

                                    
                                    for i, attack in enumerate(attacks):
                                        attack = attack["name"]
                                        plt.plot(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0), label = attack, color = colors[i], linestyle = tab_sign[i], marker = markers[i], markevery = 1)
                                        plt.fill_between(np.arange(nb_accuracies)*evaluation_delta, np.mean(tab_acc[i], axis = 0) - err[i], np.mean(tab_acc[i], axis = 0) + err[i], alpha = 0.25)

                                    plt.xlabel('Round')
                                    plt.ylabel('Accuracy')
                                    plt.xlim(0,(nb_accuracies-1)*evaluation_delta)
                                    plt.ylim(0,1)
                                    plt.grid()
                                    plt.legend()

                                    plot_name = str(dataset_name + "_"
                                                    + model_name + "_n_"
                                                    + str(nb_nodes) + "_f_"
                                                    + str(nb_byzantine) + "_"
                                                    + self.custom_dict_to_str(data_dist["name"])
                                                    + str(alpha) + "_"
                                                    + self.custom_dict_to_str(agg["name"]) + "_"
                                                    + pre_agg_names + "_lr_"
                                                    + str(lr) + "_mom_"
                                                    + str(momentum) + "_wd_"
                                                    + str(wd)+ "_lr_decay_"
                                                    + str(lr_d))
                                    plt.savefig(path_to_plot+"/"+plot_name+'_plot.pdf')
                                    plt.close()
    
    def heat_map_test_accuracy(self, path_to_results, path_to_hyperparameters, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1
        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            for agg in aggregators:
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    #try:
                    heat_map_table = np.zeros((len(alpha_list), len(nb_byz)))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, alpha in enumerate(alpha_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_"
                                                                + self.custom_dict_to_str(data_dist["name"])
                                                                + str(alpha) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                if momentum == 0.0:
                                    momentum = 0
                                wd = hyperparameters[2]
                                
                                worst_accuracy = np.inf
                                for attack in attacks:
                                    test_accuracy_file_name = str(
                                                dataset_name + "_" 
                                                + model_name + "_" 
                                                "n_" + str(nb_nodes) + "_" 
                                                + "f_" + str(nb_byzantine) + "_" 
                                                + self.custom_dict_to_str(data_dist["name"])
                                                + str(alpha) + "_" 
                                                + self.custom_dict_to_str(agg["name"]) + "_"
                                                + pre_agg_names + "_"
                                                + self.custom_dict_to_str(attack["name"]) + "_" 
                                                + "lr_" + str(lr) + "_" 
                                                + "mom_" + str(momentum) + "_" 
                                                + "wd_" + str(wd) + "_" 
                                                + "lr_decay_" + str(lr_d)
                                    )
                                    try:
                                        with open(path_to_results+ "/" + test_accuracy_file_name +'/settings.json', 'r') as file:
                                            data = json.load(file)
                                    except Exception as e:
                                        print("ERROR: "+ str(e))
                                    nb_steps = data["general"]["nb_steps"]
                                    nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
                                    tab_acc = np.zeros((
                                            nb_seeds, 
                                            nb_accuracies
                                    ))
                                    for run in range(nb_seeds):
                                        tab_acc[run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/test_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                    tab_acc = tab_acc.mean(axis=0)
                                    accuracy = np.max(tab_acc)
                                    if accuracy < worst_accuracy:
                                        worst_accuracy = accuracy
                                    
                                heat_map_table[len(heat_map_table)-1-x][y] = worst_accuracy
                    
                    file_name = str(
                        "test_"
                        + dataset_name + "_"
                        + model_name + "_"
                        + self.custom_dict_to_str(data_dist["name"]) + "_"
                        + pre_agg_names + "_"
                        + agg["name"] + ".pdf"
                    )
                    
                    column_names = [str(alpha) for alpha in alpha_list]
                    row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                    column_names.reverse()

                    sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names)
                    plt.savefig(path_to_plot +"/"+ file_name)
                    plt.close()
    
    def heat_map_validation(self, path_to_results, path_to_plot, path_to_hyperparameters):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            for agg in aggregators:
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    #try:
                    heat_map_table = np.zeros((len(alpha_list), len(nb_byz)))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, alpha in enumerate(alpha_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_"
                                                                + self.custom_dict_to_str(data_dist["name"])
                                                                + str(alpha) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                if momentum == 0.0:
                                    momentum = 0
                                wd = hyperparameters[2]
                                
                                worst_accuracy = np.inf
                                for attack in attacks:
                                    tab_acc = np.zeros((
                                            nb_seeds, 
                                            nb_accuracies
                                    ))
                                    for run in range(nb_seeds):
                                        test_accuracy_file_name = str(
                                                dataset_name + "_" 
                                                + model_name + "_" 
                                                "n_" + str(nb_nodes) + "_" 
                                                + "f_" + str(nb_byzantine) + "_" 
                                                + self.custom_dict_to_str(data_dist["name"])
                                                + str(alpha) + "_" 
                                                + self.custom_dict_to_str(agg["name"]) + "_"
                                                + pre_agg_names + "_"
                                                + self.custom_dict_to_str(attack["name"]) + "_" 
                                                + "lr_" + str(lr) + "_" 
                                                + "mom_" + str(momentum) + "_" 
                                                + "wd_" + str(wd) + "_" 
                                                + "lr_decay_" + str(lr_d) + "/"
                                            )
                                        tab_acc[run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/train_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                    tab_acc = tab_acc.mean(axis=0)
                                    accuracy = np.max(tab_acc)
                                    if accuracy < worst_accuracy:
                                        worst_accuracy = accuracy
                                    
                                heat_map_table[len(heat_map_table)-1-x][y] = worst_accuracy
                    
                    file_name = str(
                        dataset_name + "_"
                        + model_name + "_"
                        + self.custom_dict_to_str(data_dist["name"]) + "_"
                        + pre_agg_names + "_"
                        + agg["name"] + ".pdf"
                    )
                    
                    column_names = [str(alpha) for alpha in alpha_list]
                    row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                    column_names.reverse()

                    sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names)
                    plt.savefig(path_to_plot +"/"+ file_name)
                    plt.close()
                    #except Exception as e:
                    #    print(e)
    
    def heat_map_loss(self, path_to_results, path_to_hyperparameters, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1
        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            for agg in aggregators:
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    try:
                        heat_map_table = np.zeros((len(alpha_list), len(nb_byz)))
                        for nb_nodes in nb_workers:
                            for y, nb_byzantine in enumerate(nb_byz):
                                if change:
                                    nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                                for x, alpha in enumerate(alpha_list):
                                    hyperparameters_file_name = str(dataset_name + "_"
                                                                    + model_name + "_n_"
                                                                    + str(nb_nodes) + "_f_"
                                                                    + str(nb_byzantine) + "_"
                                                                    + self.custom_dict_to_str(data_dist["name"])
                                                                    + str(alpha) + "_"
                                                                    + pre_agg_names + "_"
                                                                    + agg["name"]
                                                                    + ".txt")
                                    
                                    hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                    lr = hyperparameters[0]
                                    momentum = hyperparameters[1]
                                    if momentum == 0.0:
                                        momentum = 0
                                    wd = hyperparameters[2]
                                    
                                    lowest_loss = 0
                                    for attack in attacks:
                                        test_accuracy_file_name = str(
                                                    dataset_name + "_" 
                                                    + model_name + "_" 
                                                    "n_" + str(nb_nodes) + "_" 
                                                    + "f_" + str(nb_byzantine) + "_" 
                                                    + self.custom_dict_to_str(data_dist["name"])
                                                    + str(alpha) + "_" 
                                                    + self.custom_dict_to_str(agg["name"]) + "_"
                                                    + pre_agg_names + "_"
                                                    + self.custom_dict_to_str(attack["name"]) + "_" 
                                                    + "lr_" + str(lr) + "_" 
                                                    + "mom_" + str(momentum) + "_" 
                                                    + "wd_" + str(wd) + "_" 
                                                    + "lr_decay_" + str(lr_d)
                                        )

                                        loss = np.zeros((
                                            nb_seeds, 
                                            nb_nodes,
                                            nb_steps
                                        ))
                                        for run in range(nb_seeds):
                                            for worker_i in range(data["general"]["nb_honest"]):
                                                loss[run][worker_i] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/loss_seed_"+str(run+seed)+"/loss_client_" +str(worker_i)+".txt")
                                                loss[run][worker_i] = np.nan_to_num(loss[run][worker_i], nan=np.inf)
                                        loss = np.mean(loss, axis=1)
                                        loss = np.mean(loss, axis=0)
                                        temp_lowest_loss = np.min(loss)

                                        if temp_lowest_loss > lowest_loss:
                                            lowest_loss = temp_lowest_loss
                                    heat_map_table[len(heat_map_table)-1-x][y] = lowest_loss
                        
                        file_name = str(
                            "loss_"
                            + dataset_name + "_"
                            + model_name + "_"
                            + self.custom_dict_to_str(data_dist["name"]) + "_"
                            + pre_agg_names + "_"
                            + agg["name"] + ".pdf"
                        )
                        print(file_name)
                        print(heat_map_table)
                        column_names = [str(alpha) for alpha in alpha_list]
                        row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                        column_names.reverse()
                        #sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names, cmap=sns.cm.rocket_r, vmax=0.64, vmin=0.32)
                        sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names, cmap=sns.cm.rocket_r, vmax=1.6, vmin=0.6)
                        plt.savefig(path_to_plot +"/"+ file_name)
                        plt.close()
                    
                    except Exception as e:
                        print(e)
    
    def best_heat_map_test_accuracy(self, path_to_results, path_to_hyperparameters, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1
        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            heat_map_cube = np.zeros((len(aggregators), len(data_distributions[0]["parameters"]["gamma"]), len(nb_byz)))
            for z, agg in enumerate(aggregators):
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    #try:
                    heat_map_table = np.zeros((len(alpha_list), len(nb_byz)))
                    for nb_nodes in nb_workers:
                        for y, nb_byzantine in enumerate(nb_byz):
                            if change:
                                nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                            for x, alpha in enumerate(alpha_list):
                                hyperparameters_file_name = str(dataset_name + "_"
                                                                + model_name + "_n_"
                                                                + str(nb_nodes) + "_f_"
                                                                + str(nb_byzantine) + "_"
                                                                + self.custom_dict_to_str(data_dist["name"])
                                                                + str(alpha) + "_"
                                                                + pre_agg_names + "_"
                                                                + agg["name"]
                                                                + ".txt")
                                
                                hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                lr = hyperparameters[0]
                                momentum = hyperparameters[1]
                                if momentum == 0.0:
                                    momentum = 0
                                wd = hyperparameters[2]
                                
                                worst_accuracy = np.inf
                                for attack in attacks:
                                    test_accuracy_file_name = str(
                                                dataset_name + "_" 
                                                + model_name + "_" 
                                                "n_" + str(nb_nodes) + "_" 
                                                + "f_" + str(nb_byzantine) + "_" 
                                                + self.custom_dict_to_str(data_dist["name"])
                                                + str(alpha) + "_" 
                                                + self.custom_dict_to_str(agg["name"]) + "_"
                                                + pre_agg_names + "_"
                                                + self.custom_dict_to_str(attack["name"]) + "_" 
                                                + "lr_" + str(lr) + "_" 
                                                + "mom_" + str(momentum) + "_" 
                                                + "wd_" + str(wd) + "_" 
                                                + "lr_decay_" + str(lr_d)
                                    )
                                    try:
                                        with open(path_to_results+ "/" + test_accuracy_file_name +'/settings.json', 'r') as file:
                                            data = json.load(file)
                                    except Exception as e:
                                        print("ERROR: "+ str(e))
                                    nb_steps = data["general"]["nb_steps"]
                                    nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
                                    tab_acc = np.zeros((
                                            nb_seeds, 
                                            nb_accuracies
                                    ))
                                    for run in range(nb_seeds):
                                        tab_acc[run] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/test_accuracy_seed_"+str(run+seed)+".txt", delimiter=',')
                                    tab_acc = tab_acc.mean(axis=0)
                                    accuracy = np.max(tab_acc)
                                    if accuracy < worst_accuracy:
                                        worst_accuracy = accuracy
                                    
                                heat_map_table[len(heat_map_table)-1-x][y] = worst_accuracy

                    heat_map_cube[z] = heat_map_table
                    
                file_name = str(
                    "best_test_"
                    + dataset_name + "_"
                    + model_name + "_"
                    + self.custom_dict_to_str(data_dist["name"]) + "_"
                    + pre_agg_names + ".pdf"
                )
                
                column_names = [str(alpha) for alpha in alpha_list]
                row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
                column_names.reverse()
                heat_map_table = np.max(heat_map_cube, axis=0)
                sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names)
                plt.savefig(path_to_plot +"/"+ file_name)
                plt.close()
    
    def best_heat_map_loss(self, path_to_results, path_to_hyperparameters, path_to_plot):
        try:
            with open(path_to_results+'/settings.json', 'r') as file:
                data = json.load(file)
        except Exception as e:
            print("ERROR: "+ str(e))

        if not os.path.exists(path_to_plot):
            os.makedirs(path_to_plot)
        
        seed = data["general"]["seed"]
        nb_seeds = data["general"]["nb_seeds"]
        nb_workers = data["general"]["nb_workers"]
        nb_byz = data["general"]["nb_byz"]
        nb_steps = data["general"]["nb_steps"] #Fixed
        evaluation_delta = data["general"]["evaluation_delta"] #Fixed

        model_name = data["model"]["name"] #Fixed
        dataset_name = data["model"]["dataset_name"] #Fixed
        data_distributions = data["model"]["data_distribution"]
        aggregators = data["aggregator"]
        pre_aggregators = data["pre_aggregators"]

        lr_decay = data["server"]["learning_rate_decay"]

        attacks = data["attack"]
        if isinstance(nb_workers, int):
            nb_workers = [nb_workers]
        if not isinstance(nb_workers, list):
            nb_workers = [nb_workers]
            change = True
        if isinstance(nb_byz, int):
            nb_byz = [nb_byz]
        if isinstance(data_distributions, dict):
            data_distributions = [data_distributions]
        if isinstance(aggregators, dict):
            aggregators = [aggregators]
        if isinstance(pre_aggregators[0], dict):
            pre_aggregators = [pre_aggregators]
        if isinstance(attacks, dict):
            attacks = [attacks]
        if isinstance(lr_decay, float):
            lr_decay = [lr_decay]
        
        data_folder = data["general"]["data_folder"]
        if data_folder is None:
            data_folder = "./data"
        lr_d = 0.1

        for pre_agg in pre_aggregators:
            pre_agg_list_names = [one_pre_agg['name'] for one_pre_agg in pre_agg]
            pre_agg_names = "_".join(pre_agg_list_names)
            heat_map_cube = np.zeros((len(aggregators), len(data_distributions[0]["parameters"]["gamma"]), len(nb_byz)))
            for z, agg in enumerate(aggregators):
                for data_dist in data_distributions:
                    alpha_list = [""]
                    if data_dist["name"] == "dirichlet_niid":
                        alpha_list = data_dist["parameters"]["alpha"]
                    if data_dist["name"] == "gamma_similarity_niid":
                        alpha_list = data_dist["parameters"]["gamma"]
                    try:
                        heat_map_table = np.zeros((len(alpha_list), len(nb_byz)))
                        for nb_nodes in nb_workers:
                            for y, nb_byzantine in enumerate(nb_byz):
                                if change:
                                    nb_nodes = data["general"]["nb_honest"] + nb_byzantine
                                for x, alpha in enumerate(alpha_list):
                                    hyperparameters_file_name = str(dataset_name + "_"
                                                                    + model_name + "_n_"
                                                                    + str(nb_nodes) + "_f_"
                                                                    + str(nb_byzantine) + "_"
                                                                    + self.custom_dict_to_str(data_dist["name"])
                                                                    + str(alpha) + "_"
                                                                    + pre_agg_names + "_"
                                                                    + agg["name"]
                                                                    + ".txt")
                                    
                                    hyperparameters = np.loadtxt(path_to_hyperparameters +"/hyperparameters/"+ hyperparameters_file_name)
                                    lr = hyperparameters[0]
                                    momentum = hyperparameters[1]
                                    if momentum == 0.0:
                                        momentum = 0
                                    wd = hyperparameters[2]
                                    
                                    lowest_loss = 0
                                    for attack in attacks:
                                        test_accuracy_file_name = str(
                                                    dataset_name + "_" 
                                                    + model_name + "_" 
                                                    "n_" + str(nb_nodes) + "_" 
                                                    + "f_" + str(nb_byzantine) + "_" 
                                                    + self.custom_dict_to_str(data_dist["name"])
                                                    + str(alpha) + "_" 
                                                    + self.custom_dict_to_str(agg["name"]) + "_"
                                                    + pre_agg_names + "_"
                                                    + self.custom_dict_to_str(attack["name"]) + "_" 
                                                    + "lr_" + str(lr) + "_" 
                                                    + "mom_" + str(momentum) + "_" 
                                                    + "wd_" + str(wd) + "_" 
                                                    + "lr_decay_" + str(lr_d)
                                        )
                                        """
                                        try:
                                            with open(path_to_results+ "/" + test_accuracy_file_name +'/settings.json', 'r') as file:
                                                data = json.load(file)
                                        except Exception as e:
                                            print("ERROR: "+ str(e))
                                        
                                        nb_steps = data["general"]["nb_steps"]
                                        if nb_steps == 0:
                                            nb_steps = 1
                                        
                                        nb_accuracies = int(1+math.ceil(nb_steps/evaluation_delta))
                                        """

                                        loss = np.zeros((
                                            nb_seeds, 
                                            nb_nodes,
                                            nb_steps
                                        ))
                                        for run in range(nb_seeds):
                                            for worker_i in range(data["general"]["nb_honest"]):
                                                loss[run][worker_i] = genfromtxt(path_to_results + "/" + test_accuracy_file_name+"/loss_seed_"+str(run+seed)+"/loss_client_" +str(worker_i)+".txt")
                                                loss[run][worker_i] = np.nan_to_num(loss[run][worker_i], nan=np.inf)
                                        loss = np.mean(loss, axis=1)
                                        loss = np.mean(loss, axis=0)
                                        temp_lowest_loss = np.min(loss)

                                        if temp_lowest_loss > lowest_loss:
                                            lowest_loss = temp_lowest_loss
                                    heat_map_table[len(heat_map_table)-1-x][y] = lowest_loss
                                heat_map_cube[z] = heat_map_table
                    except Exception as e:
                        print(e)
                        
            file_name = str(
                "best_loss_"
                + dataset_name + "_"
                + model_name + "_"
                + self.custom_dict_to_str(data_dist["name"]) + "_"
                + pre_agg_names + ".pdf"
            )
            column_names = [str(alpha) for alpha in alpha_list]
            row_names = [str(nb_byzantine) for nb_byzantine in nb_byz]
            column_names.reverse()
            heat_map_table = np.min(heat_map_cube, axis=0)
            sns.heatmap(heat_map_table, xticklabels=row_names, yticklabels=column_names, cmap=sns.cm.rocket_r, vmax=1.6, vmin=0.6)
            plt.savefig(path_to_plot +"/"+ file_name)
            plt.close()

plot = Plot()
"""
plot.plot_minimum_of_max_accuracies("./results", "./min_max_plots")
plot.plot_accuracy_fix_attack("./results", "./fix_attack")
plot.plot_accuracy_fix_agg("./results", "./fix_agg")
"""
#plot.plot_accuracy_fix_agg_best_setting("./results", "./best_plots", "./best_hyperparameters")
#plot.heat_map_validation("./results", "./heat_map", "./best_hyperparameters")
#plot.heat_map_test_accuracy("./results", "./best_hyperparameters", "./heat_map")
#plot.heat_map_loss("./results", "./best_hyperparameters", "./heat_map_loss")
#plot.best_heat_map_loss("./results", "./best_hyperparameters", "./heat_map_loss")
plot.best_heat_map_test_accuracy("./results", "./best_hyperparameters", "./heat_map")


from byzfl.utils.misc import check_vectors_type, random_tool
from byzfl.pipeline.robust_aggregators import RobustAggregator

class LineMaximize():
    """
    Description
    -----------
    Class to optimize attacks using the Line Maximize method: Best-effort arg-maximize a function: ℝ⁺⟶ ℝ, by mere exploration.

    Parameters
    ----------
    agg_info : dict 
        Dictionary with the keys "name" and "parameters" defined.
    pre_agg_info : list
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.
    nb_byz : int
        Number ob byzantine nodes
    evals : int
        Maximum number of evaluations, must be a positive integer
    start : float
        Initial x evaluated, must be a non-negative float
    delta : float
        Initial step delta, must be a positive float
    ratio : float
        Contraction ratio, must be between 0.5 and 1. (both excluded)

    How to use it in experiments
    ----------------------------
    >>> "attack_optimizer": {
    >>>     "name": "LineMaximize",
    >>>     "parameters": {
    >>>         "evals": 16,
    >>>         "start": 0,
    >>>         "delta": 1,
    >>>         "ratio": 0.8
    >>>     }
    >>> }

    Methods
    -------                        
    """
    def __init__(self, agg_info, pre_agg_list=[], nb_byz=0, evals=16, start=0., delta=1., ratio=0.8):
        self.robust_aggregator = RobustAggregator(
            agg_info,
            pre_agg_list
        )
        
        self.nb_byz = nb_byz
        self.evals = evals
        self.start = start
        self.delta = delta
        self.ratio = ratio
    
    def _evaluate(self,
                  attack,
                  honest_vectors,
                  avg_honest_vector,
                  attack_factor):
        """
        Compute the norm of the distance beetwen
        the difference of the honest vectors with the byzantine vectors
        aggregated against the average of the honest vectors.

        Parameters
        -----------
        attack : Attack 
            Class that simulates the attack we would like to reproduce.
            It is important that this class has the set_attack_parameters and
            get_malicious_vectors implemented. (See Fall of Empires or
            LittleIsEnough from attacks.py)

        honest_vectors : 2D ndarray or 2D torch.tensor with floating point or complex dtype
            Vectors from the honest nodes

        avg_honest_vector : 1D ndarray or 1D torch.tensor with floating point or complex dtype
            Average from the honest_vectors. 
            (Note: Could be computed inside but for questions of efficiency it's passed by parameter)
        
        attack_factor : (float)
            Attack factor to set in the attack

        Returns
        -------
        The Norm of the distance beetwen the difference of the honest vectors
        with the byzantine vectors aggregated against the average of the 
        honest vectors.
        """
        tools, honest_vectors = check_vectors_type(honest_vectors)
        
        attack.set_attack_parameters(attack_factor)

        byzantine_vector = attack(honest_vectors)

        byzantine_vectors = tools.array([byzantine_vector] * self.nb_byz)

        vectors = tools.concatenate((honest_vectors, byzantine_vectors), axis=0)
        
        agg_vectors = self.robust_aggregator(vectors)
        
        distance = tools.subtract(agg_vectors, avg_honest_vector)
        
        return tools.linalg.norm(distance)
    
    def __call__(self, attack, honest_vectors):
        """
        Iterative algorithm to set the
        best attack factor to the attack
        given the attributs of the class.

        Parameters
        -----------
        attack : Attack
            Class that simulates the attack we would like to reproduce.
            It is important that this class has the set_attack_parameters and
            get_malicious_vectors implemented. (See Fall of Empires or
            LittleIsEnough from attacks.py)

        honest_vectors : 2D ndarray or 2D torch.tensor with floating point or complex dtype
            Vectors from the honest nodes

        Note
        ----
        This function doesn't return anything but it sets to the attack
        passed by parameter the best attack factor.
        """
        tools, honest_vectors = check_vectors_type(honest_vectors)
        # Variable setup
        evals = self.evals
        delta = self.delta
        avg_honest_vector = tools.mean(honest_vectors, axis=0)
        best_x = self.start
        best_y = self._evaluate(attack, honest_vectors,
                                avg_honest_vector, best_x)
        evals -= 1
        # Expansion phase
        while evals > 0:
            prop_x = best_x + delta
            prop_y = self._evaluate(attack, honest_vectors,
                                    avg_honest_vector, prop_x)
            evals -= 1
            # Check if best
            if prop_y > best_y:
                best_y = prop_y
                best_x = prop_x
                delta *= 2
            else:
                delta *= self.ratio
                break
        # Contraction phase
        while evals > 0:
            if prop_x < best_x:
                prop_x += delta
            else:
                x = prop_x - delta
                while x < 0:
                    x = (x + prop_x) / 2
                prop_x = x
            # Same input in old doesn't correspond to same output
            # With same factor
            prop_y = self._evaluate(attack, honest_vectors,
                                    avg_honest_vector, prop_x)
            evals -= 1
            # Check if best
            if prop_y > best_y:
                best_y = prop_y
                best_x = prop_x
            # Reduce delta
            delta *= self.ratio
        # Return found maximizer
        attack.set_attack_parameters(best_x)


class WorkerWithMaxVariance():
    """
    Description
    -----------
    Class to optimize attacks that need to focus on the worker
    with maximum variance. This clase evaluates the honest vectors
    and sets the attack to focus on the worker with maximum variance.

    How to use it in experiments
    ----------------------------
    >>> "attack_optimizer": {
    >>>     "name": "WorkerWithMaxVariance",
    >>>     "parameters": {
    >>>         "steps_to_learn": 20
    >>>     }
    >>> }

    Parameters
    ----------
    steps_to_learn : int
        How many steps we are computing which worker to mimic.

    Methods
    -------          
    """
    def __init__(self, steps_to_learn=None, **kwargs):
        self.z = None
        self.mu = None
        self.steps_to_learn = steps_to_learn
        self.current_step = -1
    
    def _update_heuristic(self, honest_vectors):
        """
        Private function used to compute and update the atributs of their
        heuristic every round for computing the best z.

        Parameters
        -----------
        honest_vectors : 2D ndarray or 2D torch.tensor with floating point or complex dtype
            Vectors from the honest nodes
        """
        tools, honest_vectors = check_vectors_type(honest_vectors)
        
        if self.z is None:
            rand_tool = random_tool(honest_vectors)
            self.z = rand_tool.rand(honest_vectors[0])
        if self.mu is None:
            self.mu = tools.zeros_like(honest_vectors[0])
        
        time_factor = 1 / (self.current_step + 1)
        step_ratio = (self.current_step) * time_factor
        self.mu = tools.multiply(self.mu, step_ratio)
        self.mu = tools.add(
            self.mu,
            tools.multiply(tools.mean(honest_vectors, axis=0), time_factor)
        )

        deviations = tools.subtract(honest_vectors, self.mu)
        dot_product = tools.dot(deviations, self.z)
        dev_scaled = tools.multiply(deviations, dot_product[:,None])
        
        cumulative = tools.sum(dev_scaled, axis=0)
        cumulative = tools.divide(cumulative, tools.linalg.norm(cumulative))

        self.z = tools.multiply(self.z, step_ratio)
        self.z = tools.add(self.z, tools.multiply(cumulative, time_factor))
        self.z = tools.divide(self.z, tools.linalg.norm(self.z))
    
    def __call__(self, attack, honest_vectors):
        """
        Optimize the attack by setting their parameter to the ID 
        of the worker with more variance

        Parameters
        -----------
        attack : Attack 
            Class that simulates the attack we would like to reproduce.
            It is important that this class has the set_attack_parameters and
            get_malicious_vectors implemented.

        honest_vectors : 2D ndarray or 2D torch.tensor with floating point or complex dtype
            Vectors from the honest nodes
                
        Note
        ----
        This function doesn't return anything but it sets to the attack
        passed by parameter the ID of the worker with maximum variance.

        """
        tools, honest_vectors = check_vectors_type(honest_vectors)
        self.current_step += 1
        parameter = None

        if self.steps_to_learn is None:
            parameter = 0
        elif self.current_step < self.steps_to_learn:
            parameter = 0
            self._update_heuristic(honest_vectors)
        else:
            dot_products = abs(tools.dot(honest_vectors, self.z))
            parameter = tools.argmax(dot_products)

        attack.set_attack_parameters(parameter)




from byzfl.pipeline.robust_aggregators import RobustAggregator
from byzfl.pipeline.model_base_interface import ModelBaseInterface
from byzfl.utils.conversion import unflatten_dict

class Server(ModelBaseInterface):
    """
    Description
    -----------
    This class simulates the central server of our environment
    where the global model is updated.

    Parameters
    ----------
        All this parameters should be passed in a dictionary that contains the following keys.
    model_name : str
        Indicates the model to be used
    dataloader : Dataloader
        Dataloader with the validation set to compute the accuracy of the global model
    test_dataloader : Dataloader
        Dataloader with the validation set to compute the test accuracy of the global model
    device : str 
        Name of the device used
    bit_precision : int
        How many bits will be displayed in the accuracy
    learning_rate : float 
        Learning rate
    weight_decay : float 
        Weight decay used
    milestones : list 
        List of the milestones, where the learning rate decay should be applied
    learning-rate-decay : float
        Lerning rate decay used
    dataset_name : str 
        Name of the dataset used
    aggregator_info : dict 
        Dictionary with the keys "name" and "parameters" defined.
    pre_agg_info : list
        List of dictionaries (one for every pre_agg function)
        where every dictionary have the keys "name" and "parameters" defined.

    Methods
    -------
    """
    def __init__(self, params):
        super().__init__({
            "model_name": params["model_name"],
            "device": params["device"],
            "learning_rate": params["learning_rate"],
            "weight_decay": params["weight_decay"],
            "milestones": params["milestones"],
            "learning_rate_decay": params["learning_rate_decay"],
            "nb_workers": params["nb_workers"]
        })
        self.robust_aggregator = RobustAggregator(
            params["aggregator_info"],
            params["pre_agg_list"]
        )

        # Clipping should not be applied to batch norm stats
        batch_norm_preagg_list = []
        for pre_agg in params["pre_agg_list"]:
            if pre_agg["name"] != "Clipping":
                batch_norm_preagg_list.append(pre_agg)

        # Bath Norm Robust Aggregators
        self.robust_aggregator_mean = RobustAggregator(
            params["aggregator_info"],
            batch_norm_preagg_list
        )
        self.robust_aggregator_var = RobustAggregator(
            params["aggregator_info"],
            batch_norm_preagg_list
        )
        self.robust_aggregator_bias = RobustAggregator(
            params["aggregator_info"],
            batch_norm_preagg_list
        )

        self.model.eval()

        #Needed for batch norm
        self.batch_size = params["batch_size"]
        self.nb_workers = params["nb_workers"]
        self.batch_norm_momentum = params["batch_norm_momentum"]

    def aggregate(self, vectors):
        """
        Description
        -----------
        Aggregate vector using robust aggregation

        Parameters
        ----------
        vectors : list or np.ndarray or torch.Tensor
            A list of vectors or a matrix (2D array/tensor) where each
            row represents a vector.
        
        Returns
        -------
        list or ndarray or torch.Tensor
            The average vector of the input. The data type of the output will be the same as the input.
        """
        return self.robust_aggregator(vectors)
    
    def update_model(self, gradients):
        """
        Description
        -----------
        Update the model aggregating the gradients given and do an step.

        Parameters
        ----------
        gradients : list
            Flat list with the gradients
        """
        agg_gradients = self.aggregate(gradients)
        self.set_gradients(agg_gradients)
        self.step()
    
    def step(self):
        """
        Description
        -----------
        Do a step of the optimizer and the scheduler.
        """
        self.optimizer.step()
        self.scheduler.step()
    
    def get_model(self):
        return self.model
    
    def compute_validation_accuracy(self):
        return self._compute_accuracy(self.validation_loader)
    
    def compute_test_accuracy(self):
        return self._compute_accuracy(self.test_loader)
    
    def update_batch_norm_stats(self, new_running_mean, new_running_var):
        """
        Description
        -----------
        Update the model aggregating the bath norm statistics given.

        Parameters
        ----------
        batch_norm_stats : list
            Flat list with the bath norm statistics
        """
        agg_mean = self.robust_aggregator_mean(new_running_mean)
        agg_var = self.robust_aggregator_var(new_running_var)

        if self.use_federated_batch_norm():
            list_of_bias = [(param - agg_mean)**2 for param in new_running_mean]
            agg_bias = self.robust_aggregator_bias(list_of_bias)
            agg_var = agg_var + (self.nb_workers*self.batch_size / (self.batch_norm_momentum * (self.nb_workers*self.batch_size - 1))) * agg_bias
        
        self.set_batch_norm_stats(agg_mean, agg_var)

    def set_batch_norm_stats(self, agg_running_mean, agg_running_var):
        """
        Description
        -----------
        Sets the model batch norm stats given a flat vector.

        Parameters
        ----------
        flat_vector : list
            Flat list with the parameters
        """
        dictionary_mean, dictionary_var = self.get_batch_norm_stats()
        state_dict = self.model.state_dict()
        agg_mean_stats = unflatten_dict(dictionary_mean, agg_running_mean)
        for key, item in agg_mean_stats.items():
            state_dict[key] = item

        agg_var_stats = unflatten_dict(dictionary_var, agg_running_var)
        for key, item in agg_var_stats.items():
            state_dict[key] = item

        self.set_model_state(state_dict)



import itertools
import numpy as np
import torch
from byzfl.utils import misc

class Average(object):
    """
    Compute the average along the first axis
        
    Initialization parameters
    -------------------------
    None
        
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        >>> import byzfl
        >>> agg = byzfl.Average()

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([4. 5. 6.])
                
        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([4., 5., 6.])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([4., 5., 6.])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([4., 5., 6.])

        ---------

    """
    def __init__(self):
        pass        
    
    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        return tools.mean(vectors, axis=0)


class Median(object):

    """
    Compute the coordinate-wise median along the first axis
        

    Initialization parameters
    -------------------------
    None
        
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------

        >>> import byzfl
        >>> agg = byzfl.Median()

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([4. 5. 6.])
                
        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([4., 5., 6.])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([4., 5., 6.])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([4., 5., 6.])
    """
    def __init__(self):
        pass

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        return tools.median(vectors, axis=0)


class TrMean(object):
    r"""
    Compute the trimmed mean (or truncated mean) along the first axis:

    .. math::

        \left[\mathrm{TrMean}_{f} \ (x_1, \dots, x_n)\right]_k = \frac{1}{n - 2f}\sum_{j = f+1}^{n-f} \left[x_{\pi(j)}\right]_k
    
    where 
    
    - \\([\\cdot]_k\\) refers to the k-th coordinate

    - \\(\\pi\\) denote a permutation on \\([n]\\) that sorts the k-th
      coordinate of the input vectors in non-decreasing order, i.e., 
      \\([x_{\\pi_k(1)}]_k \\leq ...\\leq [x_{\\pi_k(n)}]_k\\)
    
    In other words, TrMean removes the \\(f\\) highest values and \\(f\\) 
    lowers values coordinate-wise, and then applies the average.

    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.

    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------

        >>> import byzfl
        >>> agg = byzfl.TrMean(1)

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([4. 5. 6.])
                
        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([4., 5., 6.])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([4., 5., 6.])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([4., 5., 6.])


    """

    def __init__(self, f=0):
        self.f = f

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        if self.f == 0:
            avg = Average()
            return avg(vectors)

        selected_vectors = tools.sort(vectors, axis=0)[self.f:-self.f]
        return tools.mean(selected_vectors, axis=0)


class GeometricMedian(object):
    r"""
    Apply the smoothed Weiszfeld algorithm [1]_ to return the
    approximate geometric median vector:

    .. math::

        \mathrm{GeometricMedian}_{\nu, T} \ (x_1, \dots, x_n) \in \argmin_{y}\sum_{i = 1}^{n} \|y - x_i\|_2
        
    
    Initialization parameters
    --------------------------
    nu : float, optional
    T : int, optional
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------

        
        >>> import byzfl
        >>> agg = byzfl.GeometricMedian()

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([3.78788764 4.78788764 5.78788764])

        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([3.7879, 4.7879, 5.7879])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([3.78788764 4.78788764 5.78788764])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([3.7879, 4.7879, 5.7879])


    References
    ----------

    .. [1] Endre Weiszfeld. Sur le point pour lequel la somme des distances de 
           n points donnés est minimum. Tohoku Mathematical Journal, First Series, 
           43:355–386, 1937

    """

    def __init__(self, nu=0.1, T=3):
        self.nu = nu
        self.T = T

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.nu, float)
        misc.check_type(self.T, int)

        z = tools.zeros_like(vectors[0])
        filtered_vectors = vectors[~tools.any(tools.isinf(vectors), axis = 1)]
        alpha = 1/len(vectors)
        for _ in range(self.T):
            betas = tools.linalg.norm(filtered_vectors - z, axis = 1)
            betas[betas<self.nu] = self.nu
            betas = (alpha/betas)[:, None]
            z = tools.sum((filtered_vectors*betas), axis=0) / tools.sum(betas)
        return z

class Krum(object):
    r"""
    Apply the Krum aggregation rule [1]_:

    .. math::

        \mathrm{Krum}_{f} \ (x_1, \dots, x_n) = x_{k^\star}
        
    with

    .. math::

        k^\star \in \argmin_{i \in [n]} \sum_{x \in \mathit{N}_i} \|x_i - x\|^2_2

    where \\(\\mathit{N}_i\\) is the set of the \\(n − f\\) nearest 
    neighbors of \\(x_i\\) in \\(\\{x_1, \\dots , x_n\\}\\)

    
    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------

        
        >>> import byzfl
        >>> agg = byzfl.Krum(1)

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([1. 2. 3.])

        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([1., 2., 3.])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray  
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([1. 2. 3.])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of  torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([1., 2., 3.])


    References
    ----------

    .. [1] Peva Blanchard, El Mahdi El Mhamdi, Rachid Guer- raoui, and Julien
           Stainer. Machine learning with adversaries: Byzantine tolerant 
           gradient descent. In I. Guyon, U. V. Luxburg, S. Bengio, H. 
           Wallach, R. Fergus, S. Vishwanathan, and R. Garnett, editors, 
           Advances in Neural Information Processing Systems 30, pages 
           119–129. Curran Associates, Inc., 2017.
    """

    def __init__(self, f=0):
        self.f = f
    
    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        distance = misc.distance_tool(vectors)

        dist = distance.cdist(vectors, vectors)**2
        dist = tools.sort(dist, axis=1)[:,1:len(vectors)-self.f]
        dist = tools.mean(dist, axis=1)
        print(dist)
        index = tools.argmin(dist)
        return vectors[index]

class MultiKrum(object):

    r"""
    Apply the Multi-Krum aggregation rule [1]_:

    .. math::

        \mathrm{MultiKrum}_{f} \ (x_1, \dots, x_n) = \frac{1}{n-f}\sum_{i = 1}^{n-f} x_{k^\star_i}
        
    with

    .. math::

        \sum_{x \in \mathit{N}_{k^\star_1} } \|x_{k^\star_1} - x\|^2_2 \leq \dots \leq \sum_{x \in \mathit{N}_{k^\star_{n-f}} } \|x_{k^\star_{n-f}} - x\|^2_2 

    where for any \\(i \\in [n], \\mathit{N}_i\\) is the set of the \\(n − f\\) nearest 
    neighbors of \\(x_i\\) in \\(\\{x_1, \\dots , x_n\\}\\)

    
    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.
    
    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.

    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.MultiKrum(1)

        Using numpy arrays
            
        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([2.5 3.5 4.5])

        Using torch tensors
            
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])

        Using list of numpy arrays

        >>> import numppy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([2.5 3.5 4.5])

        Using list of torch tensors
            
        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])


    References
    ----------

    .. [1] Peva Blanchard, El Mahdi El Mhamdi, Rachid Guerraoui, and Julien
           Stainer. Machine learning with adversaries: Byzantine tolerant 
           gradient descent. In I. Guyon, U. V. Luxburg, S. Bengio, H. 
           Wallach, R. Fergus, S. Vishwanathan, and R. Garnett, editors, 
           Advances in Neural Information Processing Systems 30, pages 
           119–129. Curran Associates, Inc., 2017.
    """

    def __init__(self, f = 0):
        self.f = f

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        distance = misc.distance_tool(vectors)
        dist = distance.cdist(vectors, vectors)**2
        dist = tools.sort(dist, axis = 1)[:,1:len(vectors)-self.f]
        dist = tools.mean(dist, axis = 1)
        k = len(vectors) - self.f
        indices = tools.argpartition(dist, k-1)[:k]
        return tools.mean(vectors[indices], axis=0)

class CenteredClipping(object):
    r"""
    Apply the Centered Clipping aggregation rule [1]_:

    .. math::

        \mathrm{CenteredClipping}_{m, L, \tau} \ (x_1, \dots, x_n) = v_{L}
        
    with

    .. math::

        v_0 &= m \\
        v_{l+1} &= v_{l} + \frac{1}{n}\sum_{i=1}^{n}(x_i - v_l)\min\left(1, \frac{\tau}{\|x_i - v_l\|}\right) \ \ ; \ \forall l \in \{0,\dots, L-1\}

    Initialization parameters
    --------------------------
    m : numpy.ndarray, torch.Tensor, optional
        Value on which the Center Clipping aggregation starts. Default makes 
        it start from a vector with all its coordinates equal to 0.
    L : int, optional
        Number of iterations. Default is set to 1.
    tau : float, optional
          Clipping threshold. Default is set to 100.

    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Note
    ----

        If the instance is called more than once, the value of \\(m\\) used in
        the next call is equal to the output vector of the previous call.

    Note
    ----
        
        In case you specify the optional parameter \\(m\\) when initializing 
        the instance, ensure that it has the same type and shape as the input
        vectors \\(x_i\\) that you will use when calling the instance.

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.CenteredClipping()

        Using numpy arrays

        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([4., 5., 6.])
        
        Using torch tensors
        
        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([4., 5., 6.])

        Using list of numpy arrays

        >>> import numpy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([4., 5., 6.])
        
        Using list of torch tensors

        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([4., 5., 6.])


    References
    ----------

    .. [1] Sai Praneeth Karimireddy, Lie He, and Martin Jaggi. Learning
           from history for byzantine robust optimization. In 38th
           International Conference on Machine Learning (ICML), 2021.
    """

    def __init__(self, m=None, L=1, tau=100):
        self.m = m
        self.L = L
        self.tau = tau
    
    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)

        if self.m is None:
            self.m = tools.zeros_like(vectors[0])

        misc.check_type(self.m, (np.ndarray, torch.Tensor))
        misc.check_type(self.L, int)
        misc.check_type(self.tau, int)

        v = self.m

        for _ in range(self.L):
            differences = vectors - v
            clip_factor = self.tau / tools.linalg.norm(differences, axis = 1)
            clip_factor = tools.minimum(tools.ones_like(clip_factor), clip_factor)
            differences = tools.multiply(differences, clip_factor.reshape(-1,1))
            v = tools.add(v, tools.mean(differences, axis=0))
        
        self.m = v

        return v


class MDA(object):
    r"""
    Apply the Minimum-Diameter Averaging aggregation rule [1]_:

    .. math::

        \mathrm{MDA}_{f} \ (x_1, \dots, x_n) = \frac{1}{n-f} \sum_{i\in S^\star} x_i
        
    with

    .. math::
        
        S^\star \in \argmin_{\substack{S \subset \{1,\dots,n\} \\ |S|=n-f}} \left\{\max_{i,j \in S} \|x_i - x_j\|_2\right\}
    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.

    Calling the instance
    --------------------

    Input parameters
    ----------------
    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.MDA(1)

        Using numpy arrays

        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([2.5, 3.5, 4.5])

        Using torch tensors

        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])

        Using list of numpy arrays

        >>> import numpy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([2.5, 3.5, 4.5])
        
        Using list of torch tensors

        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])


    References
    ----------

    .. [1] El Mhamdi, E. M., Guerraoui, R., Guirguis, A., Hoang, L. N., and 
           Rouault, S. Genuinely distributed byzantine machine learning. In 
           Proceedings of the 39th Symposium on Principles of Distributed 
           Computing, pp. 355–364, 2020.

    """

    def __init__(self, f=0):
        self.f = f        
    
    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        distance = misc.distance_tool(vectors)

        dist = distance.cdist(vectors, vectors)
        
        n = len(vectors)
        k = n - self.f

        min_diameter = np.inf
        min_subset = np.arange(k)

        all_subsets = list(itertools.combinations(range(n), k))
        for subset in all_subsets:
            vector_indices = list(itertools.combinations(subset, 2))
            diameter = tools.max(dist[tuple(zip(*vector_indices))])
            if diameter < min_diameter:
                min_subset = subset
                min_diameter = diameter
        return vectors[tools.asarray(min_subset)].mean(axis=0)


# class MVA(object):
#
#     def __init__(self, nb_byz):
#         self.nb_byz = nb_byz
#
#     def __call__(self, vectors):
#         tools, vectors = misc.check_vectors_type(vectors)
#         misc.check_type(self.nb_byz, int)
#
#         distance = misc.distance_tool(vectors)
#
#         dist = distance.cdist(vectors, vectors)**2
#
#         n = len(vectors)
#         k = n - self.nb_byz
#
#         min_diameter = np.inf
#         min_subset = np.arange(k)
#
#         all_subsets = list(itertools.combinations(range(n), k))
#         for subset in all_subsets:
#             vector_indices = list(itertools.combinations(subset, 2))
#             diameter = tools.sum(dist[tuple(zip(*vector_indices))])
#             if diameter < min_diameter:
#                 min_subset = subset
#                 min_diameter = diameter
#
#         return vectors[tools.asarray(min_subset)].mean(axis=0)

class MoNNA(object):


    r"""
    Apply the MoNNA rule [1]_:

    .. math::

        \mathrm{MoNNA}_{f, \mathrm{idx}} \ (x_1, \dots, x_n) = \frac{1}{n-f} \sum_{i \in \mathit{N}_{\mathrm{idx}}} x_{i}
        
    where \\(\\mathit{N}_{\\mathrm{idx}}\\) is the set of the \\(n − f\\) nearest 
    neighbors of \\(x_{\\mathrm{idx}}\\) in \\(\\{x_1, \\dots , x_n\\}\\)



    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.
    idx : int, optional
        Index of the vector on which the neighborhood will be computed. The 
        default is setting \\(\\mathrm{idx}=0\\).

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Note
    ----

    MoNNA is used in peer-to-peer settings where the idx refer to a vectors 
    that is known to be correct (i.e. not faulty).

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.MoNNA(1, 1)

        Using numpy arrays

        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([2.5, 3.5, 4.5])

        Using torch tensors

        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])

        Using list of numpy arrays

        >>> import numpy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([2.5, 3.5, 4.5])
        
        Using list of torch tensors

        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])


    References
    ----------

    .. [1] Farhadkhani, S., Guerraoui, R., Gupta, N., Hoang, L. N., Pinot, R.,
           & Stephan, J. (2023, July). Robust collaborative learning with 
           linear gradient overhead. In International Conference on Machine 
           Learning (pp. 9761-9813). PMLR. 

    """

    
    def __init__(self, f=0, idx=0):
        self.f = f
        self.idx = idx
    
    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)
        misc.check_type(self.idx, int)

        distance = misc.distance_tool(vectors)

        dist = distance.cdist(vectors, vectors[self.idx].reshape(1,-1))
        k = len(vectors) - self.f
        indices = tools.argpartition(dist.reshape(-1), k-1)[:k]
        return tools.mean(vectors[indices], axis=0)


class Meamed(object):

    r"""

    .. math::

        

    Initialization parameters
    --------------------------
    f : int, optional
        Number of faulty vectors. Set to 0 by default.

    Calling the instance
    --------------------

    Input parameters
    ----------------

    vectors: numpy.ndarray, torch.Tensor, list of numpy.ndarray or list of torch.Tensor
        A set of vectors, matrix or tensors.
        
    Returns
    -------
    :numpy.ndarray or torch.Tensor
        The data type of the output will be the same as the input.

    Examples
    --------
        
        >>> import byzfl
        >>> agg = byzfl.Meamed(1)

        Using numpy arrays

        >>> import numpy as np
        >>> x = np.array([[1., 2., 3.],       # np.ndarray
        >>>               [4., 5., 6.], 
        >>>               [7., 8., 9.]])
        >>> agg(x)
        array([2.5, 3.5, 4.5])

        Using torch tensors

        >>> import torch
        >>> x = torch.tensor([[1., 2., 3.],   # torch.tensor 
        >>>                   [4., 5., 6.], 
        >>>                   [7., 8., 9.]])
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])

        Using list of numpy arrays

        >>> import numpy as np
        >>> x = [np.array([1., 2., 3.]),      # list of np.ndarray 
        >>>      np.array([4., 5., 6.]), 
        >>>      np.array([7., 8., 9.])]
        >>> agg(x)
        array([2.5, 3.5, 4.5])
        
        Using list of torch tensors

        >>> import torch
        >>> x = [torch.tensor([1., 2., 3.]),  # list of torch.tensor 
        >>>      torch.tensor([4., 5., 6.]), 
        >>>      torch.tensor([7., 8., 9.])]
        >>> agg(x)
        tensor([2.5000, 3.5000, 4.5000])


    References
    ----------

    .. [1] XXXXX

    """
    def __init__(self, f=0):
        self.f = f

    def __call__(self, vectors):
        tools, vectors = misc.check_vectors_type(vectors)
        misc.check_type(self.f, int)

        d = len(vectors[0])
        k = len(vectors) - self.f

        median = tools.median(vectors, axis=0)
        abs_diff = tools.abs((vectors - median))

        indices = tools.argpartition(abs_diff, k-1, axis=0)[:k]
        indices = tools.multiply(indices, d)
        a = tools.arange(d)
        if not tools == np:
            a = a.to(indices.device)
        indices = tools.add(indices, a)
        return tools.mean(vectors.take(indices), axis=0)